/*
 * Copyright (c) 2017, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

import UIKit
import CoreBluetooth

class BLEDeviceTableViewController: UITableViewController,
    CBCentralManagerDelegate
{
    // MARK: Properties:
    @IBOutlet var deviceTable: UITableView!
    
    // MARK: Variables
    var btManager: CBCentralManager!
    var btDevices: [CBPeripheral]! = []
    var selBTDev: CBPeripheral!
    var rssiReadings: [NSNumber]! = []
    var advData: [NSData]! = []
    var goSegueIdentifier = ""
    let identifier = "BLEDeviceTableViewCell"
    var targetName = ""
    var codeExampleTable: UIViewController!
    var connectionCon: DiscoveryViewController!
    var timeoutTimer: Timer!
    var advertisementData: NSData!
    
    override func viewDidLoad()
    {
        super.viewDidLoad()
        btManager = CBCentralManager(delegate: self, queue: nil)
    }

    override func didReceiveMemoryWarning()
    {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }

    // MARK: - Table view data source
    override func numberOfSections(in tableView: UITableView) -> Int
    {
        return 1
    }

    override func tableView(_ tableView: UITableView,
                            numberOfRowsInSection section: Int) -> Int
    {
        return btDevices.count
    }

    override func tableView(_ tableView: UITableView,
                            cellForRowAt indexPath: IndexPath) -> UITableViewCell
    {
        let cell: BLEDeviceTableViewCell = tableView.dequeueReusableCell(withIdentifier: identifier, for:
                    indexPath) as! BLEDeviceTableViewCell
        let device = btDevices[indexPath.row]
        let devName = device.name

        if devName != nil
        {
            cell.deviceText.text = device.name
        }
        else
        {
            cell.deviceText.text = "Unnamed Device"
        }
        
        cell.setSignalStrengthImage(rssi: rssiReadings[indexPath.row].floatValue)

        return cell
    }
    
    override func tableView(_ tableView: UITableView,
                            didSelectRowAt indexPath: IndexPath)
    {
        connectionCon = DiscoveryViewController(message: "Connecting...")
        present(connectionCon, animated: true, completion: nil)
        
        timeoutTimer = Timer.scheduledTimer(timeInterval: 5.0,
                                            target: self,
                                            selector: #selector(self.completionTimeout),
                                            userInfo: nil,
                                            repeats: false)
        
        selBTDev = btDevices[indexPath.row]
        btManager.connect(selBTDev, options: nil)
    }
    
    // MARK: CBCentralManagerDelegate
    func centralManagerDidUpdateState(_ central: CBCentralManager)
    {
        if central.state == .poweredOn
        {
            central.scanForPeripherals(withServices: nil, options: nil)
        }
        else
        {
            _ = self.navigationController?.popViewController(animated: true);
        }
    }
    
    func centralManager(_ central: CBCentralManager,
                        didDiscover peripheral: CBPeripheral,
                        advertisementData: [String : Any],
                        rssi RSSI: NSNumber)
    {
        if(!btDevices.contains(peripheral))
        {
            var insertRow = 0
            
            
            while(insertRow < rssiReadings.count)
            {
                if(rssiReadings[insertRow].doubleValue < RSSI.doubleValue)
                {
                    break
                }
                
                insertRow = insertRow + 1
            }
            
            let newIndexPath = IndexPath(row: insertRow, section: 0)
            btDevices.insert(peripheral, at: insertRow)
            rssiReadings.insert(RSSI, at: insertRow)
            
            if let manData = advertisementData[CBAdvertisementDataManufacturerDataKey] as? NSData
            {
                advData.insert(manData, at: insertRow)
            }
            else
            {
                advData.insert(NSData(), at: insertRow)
            }
            
            tableView.insertRows(at: [newIndexPath], with: .bottom)
        }
    }
    
    func centralManager(_ central: CBCentralManager,
                        didConnect peripheral: CBPeripheral)
    {
        self.performSegue(withIdentifier: goSegueIdentifier, sender: self)
    }
    
    func centralManager(_ central: CBCentralManager,
                        didDisconnectPeripheral peripheral: CBPeripheral,
                        error: Error?)
    {
        if(self.btManager.isScanning)
        {
            return
        }
        
        let viewControllers = self.navigationController?.viewControllers
        
        if let curController = viewControllers?[(viewControllers?.count)!-1] as? OADViewController
        {
            if(curController.firmwareDownloadView != nil && curController.firmwareDownloadView.isViewLoaded)
            {
                curController.firmwareDownloadView.dismiss(animated: false, completion: displayDisconnect)
                return
            }
        }
        
        displayDisconnect()
        
    }
    
    func displayDisconnect()
    {
        let alert = UIAlertController(title: "Disconnected",
                                      message: "Device was disconnected",
                                      preferredStyle: UIAlertControllerStyle.alert)
        alert.addAction(UIAlertAction(title: "OK",
                                      style: UIAlertActionStyle.default,
                                      handler: self.completionDisconnect))
        self.codeExampleTable.present(alert, animated: true,
                                      completion: nil)
    }
    
    func completionDisconnect(action: UIAlertAction)
    {
        _ = self.navigationController?.popToViewController(codeExampleTable, animated: true)
    }
    
    @objc func completionTimeout()
    {
        btManager.cancelPeripheralConnection(selBTDev)
        connectionCon.dismiss(animated: true, completion: nil)
        
        let alert = UIAlertController(title: "Connection Timeout",
                                      message: "Timeout occurred while connecting to device. Please verify correct" +
                                                " services are being advertised on device.",
                                      preferredStyle: UIAlertControllerStyle.alert)
        alert.addAction(UIAlertAction(title: "OK",
                                      style: UIAlertActionStyle.default,
                                      handler: nil))
        _ = navigationController?.popToViewController(self, animated: true)
        present(alert, animated: true, completion: nil)
    }

    // MARK: - Navigation
    override func prepare(for segue: UIStoryboardSegue, sender: Any?)
    {
        if(segue.identifier == "projectZeroSegue")
        {
            let projZeroViewCon: ProjectZeroTabBarController =
                segue.destination as! ProjectZeroTabBarController
            projZeroViewCon.btDevice = selBTDev
            selBTDev.delegate = projZeroViewCon
            selBTDev.discoverServices(nil)
            projZeroViewCon.referencedController  = self
        }
        else if(segue.identifier == "oadSegue")
        {
            let oadViewController: OADViewController =
                segue.destination as! OADViewController
            oadViewController.btDevice = selBTDev
            oadViewController.codeExampleTable = self.codeExampleTable
            oadViewController.referencedController = self
        }
        else if(segue.identifier == "tempSensorSegue")
        {
            let tempSensorViewController: TemperatureSensorViewController =
                segue.destination as! TemperatureSensorViewController
            tempSensorViewController.btDevice = selBTDev
            selBTDev.delegate = tempSensorViewController
            selBTDev.discoverServices(nil)
            tempSensorViewController.referencedController = self
        }
        else if(segue.identifier == "sapSegue")
        {
            let sapViewController: SimpleAPViewController =
                segue.destination as! SimpleAPViewController
            sapViewController.btDevice = selBTDev
            selBTDev.delegate = sapViewController
            selBTDev.discoverServices(nil)
            sapViewController.referencedController = self
        }
        else if(segue.identifier == "sensorSegue")
        {
            let sensorViewCon: SensorTabBarController =
                segue.destination as! SensorTabBarController
            sensorViewCon.btDevice = selBTDev
            selBTDev.delegate = sensorViewCon
            selBTDev.discoverServices(nil)
            sensorViewCon.referencedController  = self
        }
        else if(segue.identifier == "eLockSegue")
        {
            let eLockCon: LockViewController =
                segue.destination as! LockViewController
            eLockCon.btDevice = selBTDev
            selBTDev.delegate = eLockCon
            selBTDev.discoverServices(nil)
            eLockCon.referencedController  = self
        }
        
        btManager.stopScan()
        btDevices.removeAll()
        rssiReadings.removeAll()
        advData.removeAll()
        tableView.reloadData()
    }
}
