/*
 * Copyright (c) 2018, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
import UIKit
import CoreBluetooth

class BLEProvisionDeviceViewController: UIViewController, UITableViewDelegate, UITableViewDataSource,
    CBCentralManagerDelegate
{
    // MARK: Properties
    @IBOutlet weak var tableView: UITableView!
    @IBOutlet weak var dirLabel: UILabel!
    
    // MARK: Variables
    var btManager: CBCentralManager!
    var btDevices: [CBPeripheral]! = []
    var selBTDev: CBPeripheral!
    var rssiReadings: [NSNumber]! = []
    let identifier = "BLEProvisionDeviceCell"
    var provisioningController: UIViewController!
    var routerInfoController: BLEProvisionInfoViewController!
    var connectionCon: DiscoveryViewController!
    var timeoutTimer: Timer!
    var unwindingFromRouter: Bool = false
    let serviceUUIDs: [CBUUID] = [BLE_WIFI_SERVICE_UUID]

    override func viewDidLoad()
    {
        super.viewDidLoad()
        btManager = CBCentralManager(delegate: self, queue: nil)
        tableView.dataSource = self
        tableView.delegate = self
        dirLabel.adjustsFontSizeToFitWidth = true
        unwindingFromRouter = false
        
        NotificationCenter.default.addObserver(self,
                                               selector: #selector(self.viewRotated),
                                               name: NSNotification.Name.UIDeviceOrientationDidChange, object: nil)
        
        viewRotated()
    }
    
    // MARK: - Table view data source
    func numberOfSections(in tableView: UITableView) -> Int
    {
        return 1
    }
    
    func tableView(_ tableView: UITableView,
                   numberOfRowsInSection section: Int) -> Int
    {
        return btDevices.count
    }
    
    func tableView(_ tableView: UITableView, heightForRowAt indexPath: IndexPath) -> CGFloat
    {
        return 40.0
    }
    
    func tableView(_ tableView: UITableView,
                   cellForRowAt indexPath: IndexPath) -> UITableViewCell
    {
        let cell: BLEProvisionTableViewCell = tableView.dequeueReusableCell(withIdentifier: identifier, for:
            indexPath) as! BLEProvisionTableViewCell
        let device = btDevices[indexPath.row]
        let devName = device.name
        
        if devName != nil
        {
            cell.devLabel.text = device.name
        }
        else
        {
            cell.devLabel.text = "Unnamed Device"
        }
        
        cell.setSignalStrengthImage(rssi: rssiReadings[indexPath.row].floatValue)
        
        return cell
    }
    
    func tableView(_ tableView: UITableView,
                   didSelectRowAt indexPath: IndexPath)
    {
        selBTDev = btDevices[indexPath.row]
        
        connectionCon = DiscoveryViewController(message: "Connecting...")
        present(connectionCon, animated: true, completion: nil)
        timeoutTimer = Timer.scheduledTimer(timeInterval: 5.0,
                                            target: self,
                                            selector: #selector(self.completionTimeout),
                                            userInfo: nil,
                                            repeats: false)
        btManager.connect(selBTDev, options: nil)
    }
    
    // MARK: CBCentralManagerDelegate
    func centralManagerDidUpdateState(_ central: CBCentralManager)
    {
        if central.state == .poweredOn
        {
            central.scanForPeripherals(withServices: serviceUUIDs, options: nil)
        }
        else
        {
            _ = self.navigationController?.popViewController(animated: true);
        }
    }
    
    func centralManager(_ central: CBCentralManager,
                        didConnect peripheral: CBPeripheral)
    {
        if(routerInfoController == nil)
        {
            self.performSegue(withIdentifier: "routerSegue", sender: self)
        }
        else if(routerInfoController.confirmationInProgress == true)
        {
            routerInfoController.btDevice = peripheral
            routerInfoController.finishConfirmation()
        }
    }
    
    func centralManager(_ central: CBCentralManager,
                        didDiscover peripheral: CBPeripheral,
                        advertisementData: [String : Any],
                        rssi RSSI: NSNumber)
    {
        if(routerInfoController != nil)
        {
            return
        }
        
        if(!btDevices.contains(peripheral))
        {
            var insertRow = 0
            
            while(insertRow < rssiReadings.count)
            {
                if(rssiReadings[insertRow].doubleValue < RSSI.doubleValue)
                {
                    break
                }
                
                insertRow = insertRow + 1
            }
            
            let newIndexPath = IndexPath(row: insertRow, section: 0)
            btDevices.insert(peripheral, at: insertRow)
            rssiReadings.insert(RSSI, at: insertRow)
            tableView.insertRows(at: [newIndexPath], with: .bottom)
        }
    }
    
    func centralManager(_ central: CBCentralManager,
                        didDisconnectPeripheral peripheral: CBPeripheral,
                        error: Error?)
    {
        if((routerInfoController != nil) && ((routerInfoController.confirmationInProgress == true) ||
            (routerInfoController.provisioningComplete == true)))
        {
            return
        }
        
        if(unwindingFromRouter == true)
        {
            unwindingFromRouter = false
            return
        }
        
        displayDisconnect()
    }
    
    func displayDisconnect()
    {
        let alert = UIAlertController(title: "Disconnected",
                                      message: "Device was disconnected",
                                      preferredStyle: UIAlertControllerStyle.alert)
        alert.addAction(UIAlertAction(title: "OK",
                                      style: UIAlertActionStyle.default,
                                      handler: self.completionDisconnect))
        self.present(alert, animated: true, completion: nil)
    }
    
    func completionDisconnect(action: UIAlertAction)
    {
        _ = self.navigationController?.popToViewController(self, animated: true)
    }
    
    @objc func completionTimeout()
    {
        btManager.cancelPeripheralConnection(selBTDev)
        connectionCon.dismiss(animated: false, completion: nil)
        
        let alert = UIAlertController(title: "Connection Timeout",
                                      message: "Timeout occurred while connecting to device. Please verify correct" +
            " services are being advertised on device.",
                                      preferredStyle: UIAlertControllerStyle.alert)
        alert.addAction(UIAlertAction(title: "OK",
                                      style: UIAlertActionStyle.default,
                                      handler: nil))
        _ = navigationController?.popToViewController(self, animated: true)
        present(alert, animated: true, completion: nil)
    }
    
    // MARK: - Navigation
    override func prepare(for segue: UIStoryboardSegue, sender: Any?)
    {
        /* Setting references in the device info controller to all of the relevant BLE variables */
        let bleDevController: BLEProvisionInfoViewController = segue.destination as! BLEProvisionInfoViewController
        routerInfoController = bleDevController
        bleDevController.referenceController =  self
        bleDevController.btDevice = selBTDev
        selBTDev.delegate = bleDevController
        
        /* Starting to discover services */
        selBTDev.discoverServices(nil)
        
        /* Removing all of the devices  and stopping the active scan*/
        btManager.stopScan()
        btDevices.removeAll()
        rssiReadings.removeAll()
        tableView.reloadData()
    }
    
    override func viewWillAppear(_ animated: Bool)
    {
        /* Scanning for peripherals */
        if(btManager != nil && btManager.state == .poweredOn)
        {
            btManager.scanForPeripherals(withServices: serviceUUIDs, options: nil)
        }
        
        if(routerInfoController != nil)
        {
            btManager.cancelPeripheralConnection(selBTDev)
            unwindingFromRouter = true
        }
        
        routerInfoController = nil
    }
    
    @objc func viewRotated()
    {
        DispatchQueue.main.async
            {
                if UIDeviceOrientationIsPortrait(UIDevice.current.orientation)
                {
                    self.dirLabel.isHidden = false
                }
                
                if UIDeviceOrientationIsLandscape(UIDevice.current.orientation)
                {
                    self.dirLabel.isHidden = true
                }
        }
    }
}
