/*
 * Copyright (c) 2018, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
import UIKit
import CoreBluetooth

class BLEProvisionInfoViewController: UIViewController, UITextFieldDelegate, CBPeripheralDelegate
{
    // MARK: Properties
    @IBOutlet weak var routerImage: UIImageView!
    @IBOutlet weak var routerLabel: UILabel!
    @IBOutlet weak var ssidField: UITextField!
    @IBOutlet weak var keyField: UITextField!
    @IBOutlet weak var devField: UITextField!
    @IBOutlet weak var provButton: UIButton!
    @IBOutlet weak var bottomMargin: NSLayoutConstraint!
    
    // MARK: Variables
    var statusChar: CBCharacteristic!
    var notifyChar: CBCharacteristic!
    var ssidChar: CBCharacteristic!
    var passChar: CBCharacteristic!
    var devChar: CBCharacteristic!
    var startChar: CBCharacteristic!
    var btDevice: CBPeripheral!
    var referenceController: BLEProvisionDeviceViewController!
    var provisionController: BLEProvisionViewController!
    var confirmationTimer: Timer!
    var confirmationInProgress: Bool = false
    var provisioningComplete: Bool = false
    
    // MARK: Actions
    @IBAction func provisionWasPressed(_ sender: Any)
    {
        if(ssidField.text?.isEmpty == true)
        {
            let alert = UIAlertController(title: "SSID is blank",
                                          message: "Value must be specified for SSID of router!",
                                          preferredStyle: UIAlertControllerStyle.alert)
            alert.addAction(UIAlertAction(title: "OK", style: UIAlertActionStyle.default, handler: nil))
            self.present(alert, animated: true, completion: nil)
            return
        }
        
        if(devField.text?.isEmpty == true)
        {
            devField.text = "mysimplelink"
        }
        
        /* Move on to provision */
        performSegue(withIdentifier: "provisionSegue", sender: self)
    }
    
    // MARK: Functions
    override func viewDidLoad()
    {
        super.viewDidLoad()
        
        /* Adjusting the GUI */
        ssidField.layer.cornerRadius = 8.0
        keyField.layer.cornerRadius = 8.0
        devField.layer.cornerRadius = 8.0
        provButton.layer.cornerRadius = 8.0
        ssidField.clipsToBounds = false
        devField.clipsToBounds = false
        keyField.clipsToBounds = false
        provButton.clipsToBounds = false
        
        /* Setting the delegates */
        ssidField.delegate = self
        keyField.delegate = self
        devField.delegate = self
        
        NotificationCenter.default.addObserver(self,
                                               selector: #selector(BLEProvisionInfoViewController.viewRotated),
                                               name: NSNotification.Name.UIDeviceOrientationDidChange, object: nil)
        
        /* Adding observer to shift view for keyboard */
        NotificationCenter.default.addObserver(self,
                                               selector: #selector(self.keyboardNotification),
                                               name: NSNotification.Name.UIKeyboardWillChangeFrame,
                                               object: nil)
        
        viewRotated()
    }
    
    func finishConfirmation()
    {
        /* Reading the status bit */
        provisionController.printLogMessage(message: "Device reconnected.")
        provisionController.printLogMessage(message: "Reading confirmation status value.")
        btDevice.discoverServices(nil)
    }
    
    // MARK: CBPeripheralDelegate
    func peripheral(_ peripheral: CBPeripheral,
                    didDiscoverServices error: Error?)
    {
        for service in peripheral.services!
        {
            if(service.uuid == BLE_WIFI_SERVICE_UUID)
            {
                let thisService = service as CBService
                peripheral.discoverCharacteristics(nil, for: thisService)
                return
            }
        }
    }
    
    func peripheral(_ peripheral: CBPeripheral,
                    didDiscoverCharacteristicsFor service: CBService,
                    error: Error?)
    {
        /* Setting all of the expected characteristics */
        for characteristic in service.characteristics!
        {
            if(characteristic.uuid == BLE_WIFI_STATUS_UUID)
            {
                self.statusChar = characteristic
            }
            else if(characteristic.uuid == BLE_WIFI_NOTIFY_UUID)
            {
                self.notifyChar = characteristic
            }
            else if(characteristic.uuid == BLE_WIFI_SSID_UUID)
            {
                self.ssidChar = characteristic
            }
            else if(characteristic.uuid == BLE_WIFI_PASS_UUID)
            {
                self.passChar = characteristic
            }
            else if(characteristic.uuid == BLE_WIFI_DEVNAME_UUID)
            {
                self.devChar = characteristic
            }
            else if(characteristic.uuid == BLE_WIFI_START_UUID)
            {
                self.startChar = characteristic
            }
        }
        
        /* If all the characteristics are not nil it means that we found all of our expected characteristics */
        if((notifyChar != nil) && (statusChar != nil) && (ssidChar != nil) && (passChar != nil) && (devChar != nil)
            && (startChar != nil))
        {
            if(confirmationInProgress == false)
            {
                referenceController.connectionCon.dismiss(animated: true, completion: nil)
                referenceController.timeoutTimer.invalidate()
            }
            else
            {
                btDevice.readValue(for: statusChar)
            }
        }
    }
    
    /* Callback for a characteristic write */
    func peripheral(_ peripheral: CBPeripheral, didWriteValueFor characteristic: CBCharacteristic, error: Error?)
    {
        if(characteristic.uuid == BLE_WIFI_SSID_UUID)
        {
            provisionController.printLogMessage(message: "SSID characteristic written successfully!")
            
            if((keyField.text?.isEmpty)! == false)
            {
                provisionController.printLogMessage(message: "Writing security key characteristic.")
                self.btDevice.writeValue((keyField.text?.data(using: String.Encoding.utf8))!, for: passChar,
                                     type: CBCharacteristicWriteType.withResponse)
                return
            }
        }
        else if(characteristic.uuid == BLE_WIFI_PASS_UUID)
        {
            provisionController.printLogMessage(message: "Security key characteristic written successfully!")
            
            if((devField.text?.isEmpty)! == false)
            {
                provisionController.printLogMessage(message: "Writing device name characteristic.")
                self.btDevice.writeValue((ssidField.text?.data(using: String.Encoding.utf8))!, for: devChar,
                                         type: CBCharacteristicWriteType.withResponse)
                return
            }
        }
        else if(characteristic.uuid == BLE_WIFI_DEVNAME_UUID)
        {
            provisionController.printLogMessage(message: "Device name characteristic written successfully!")
        }
        else if(characteristic.uuid == BLE_WIFI_START_UUID)
        {
            provisionController.printLogMessage(message: "Start characteristic written successfully!")
            
            if(provisioningComplete == true)
            {
                provisionController.printLogMessage(message: "Provisioning successfully completed.")
                provisionController.statusLabel.text = "Complete"
                provisionController.statusLabel.textColor = UIColor(red: 78.0/255.0, green: 171.0/255.0,
                                                                    blue: 17.0/255.0, alpha: 1)
                provisionController.completeButton.isEnabled = true
                DispatchQueue.main.async
                {
                    self.provisionController.circleView.layer.removeAllAnimations()
                }
                
                let dateFormatter = DateFormatter()
                dateFormatter.dateStyle = .long
                dateFormatter.timeStyle = .short
                
                /* Filling out the provisioned device structure and adding it to the provisioned devices */
                let curProvisionedDevice = ProvisionedWiFiDevice()
                var wifiSettings = WiFiSettings()
                
                if(devField.text?.isEmpty == true)
                {
                    wifiSettings.devName = "mysimplelink"
                }
                else
                {
                    wifiSettings.devName = devField.text!
                }
                
                wifiSettings.routerName = ssidField.text!
                
                if(keyField.text?.isEmpty == true)
                {
                    wifiSettings.routerKey = "N/A"
                    wifiSettings.routerSecType = .None
                }
                else
                {
                    wifiSettings.routerKey = keyField.text!
                    wifiSettings.routerSecType = .WPA
                }
                
                wifiSettings.simplelinkKey = ""
                wifiSettings.simplelinkName = ""
                wifiSettings.devSecType = .None
                
                curProvisionedDevice.confirmMethod = "BLE Confirmation"
                curProvisionedDevice.dateAdded = dateFormatter.string(from: Date())
                curProvisionedDevice.provisionLog = provisionController.logView.text
                curProvisionedDevice.commonURL = ""
                curProvisionedDevice.devURL = ""
                curProvisionedDevice.devVersion = ""
                curProvisionedDevice.directURL = ""
                
                curProvisionedDevice.wifiSettings = wifiSettings
                
                var provisionedDevices = loadProvisionedDevices()!
                provisionedDevices.append(curProvisionedDevice)
                NSKeyedArchiver.archiveRootObject(provisionedDevices, toFile: ProvisionedWiFiDevice.ArchiveURL.path)
            }
            
            return
        }
        else
        {
            provisionController.printLogMessage(message: "WARNING: Unknown characteristic write!!")
            return
        }
    
        let writeArray = Data([UInt8]([0x01]))
        
        provisionController.printLogMessage(message: "Initiating provisioning. This will cause the device to" +
                        " disconnect and for the SimpleLink device to attempt to connect to the configured router")
        self.btDevice.writeValue(writeArray, for: startChar, type: CBCharacteristicWriteType.withResponse)
        
        provisionController.statusLabel.text = "Provisioning"
        
        confirmationInProgress = true
        confirmationTimer = Timer.scheduledTimer(timeInterval: 15.0,
                                            target: self,
                                            selector: #selector(self.completionTimeout),
                                            userInfo: nil,
                                            repeats: false)
    }
    
    /* Callback for reading a characteristic */
    func peripheral(_ peripheral: CBPeripheral, didUpdateValueFor characteristic: CBCharacteristic, error: Error?)
    {
        /* Reading the status bit */
        if(characteristic.uuid == BLE_WIFI_STATUS_UUID)
        {
            var values = [UInt8](repeating:0, count:characteristic.value!.count)
            var intValue: UInt8 = 0
            characteristic.value?.copyBytes(to: &values,
                                            count: characteristic.value!.count)
            memcpy(&intValue, values, 1)
            
            if(intValue == WIFI_BLE_PROVSTAT_SUCCESS)
            {
                provisionController.printLogMessage(message: "IP Acquired and provisioning succeeded." +
                                                        " Sending confirmation")
                provisionController.statusLabel.text = "Confirming"
                confirmationInProgress = false
                provisioningComplete = true
                let writeArray = Data([UInt8]([0x02]))
                btDevice.writeValue(writeArray, for: startChar, type: CBCharacteristicWriteType.withResponse)
                return
            }
            else
            {
                if(intValue == WIFI_BLE_PROVSTAT_FAILCON)
                {
                    provisionController.printLogMessage(message: "ERROR: Connection to AP failed!")
                }
                else if(intValue == WIFI_BLE_PROVSTAT_TIMEOUT)
                {
                    provisionController.printLogMessage(message: "ERROR: Provisioning timeout!")
                }
                else if(intValue == WIFI_BLE_PROVSTAT_FAILPING)
                {
                    provisionController.printLogMessage(message: "ERROR: Failed to ping gateway!")
                }
                else if(intValue == WIFI_BLE_PROVSTAT_FAILIP)
                {
                    provisionController.printLogMessage(message: "ERROR: Failed to acquire IP!")
                }
                
                provisionController.statusLabel.text = "Failed"
                provisionController.statusLabel.textColor = UIColor.red
                
                DispatchQueue.main.async
                {
                    self.provisionController.circleView.layer.removeAllAnimations()
                }
            }
        }
    }
    
    // MARK: - UITextFieldDelegate
    func textFieldShouldReturn(_ textField: UITextField) -> Bool
    {
        if(textField == ssidField)
        {
            ssidField.resignFirstResponder()
        }
        else if(textField == keyField)
        {
            keyField.resignFirstResponder()
        }
        else if(textField == devField)
        {
            devField.resignFirstResponder()
        }
        
        routerImage.isHidden = false
        
        return true
    }
    
    func textFieldDidBeginEditing(_ textField: UITextField)
    {
        routerImage.isHidden = true
    }
    
    func textFieldDidEndEditing(_ textField: UITextField)
    {
        
    }
    override func prepare(for segue: UIStoryboardSegue, sender: Any?)
    {
        let bleProvCon: BLEProvisionViewController = segue.destination as! BLEProvisionViewController
        self.provisionController = bleProvCon
        bleProvCon.referenceController =  self
    }
    
    func startProvisioning()
    {
        provisionController.printLogMessage(message: "Writing SSID characteristic (" + ssidField.text! + ")")
        self.btDevice.writeValue((ssidField.text?.data(using: String.Encoding.utf8))!, for: ssidChar,
                                 type: CBCharacteristicWriteType.withResponse)
    }
    
    @objc func keyboardNotification(notification: NSNotification)
    {
        if let userInfo = notification.userInfo
        {
            let endFrame = (userInfo[UIKeyboardFrameEndUserInfoKey] as? NSValue)?.cgRectValue
            let duration:TimeInterval = (userInfo[UIKeyboardAnimationDurationUserInfoKey] as? NSNumber)?.doubleValue ?? 0
            let animationCurveRawNSN = userInfo[UIKeyboardAnimationCurveUserInfoKey] as? NSNumber
            let animationCurveRaw = animationCurveRawNSN?.uintValue ?? UIViewAnimationOptions.curveEaseInOut.rawValue
            let animationCurve:UIViewAnimationOptions = UIViewAnimationOptions(rawValue: animationCurveRaw)
            if (endFrame?.origin.y)! >= UIScreen.main.bounds.size.height
            {
                bottomMargin?.constant = 10.0
            }
            else
            {
                bottomMargin?.constant = endFrame?.size.height ?? 10.0
            }
            UIView.animate(withDuration: duration,
                           delay: TimeInterval(0),
                           options: animationCurve,
                           animations: { self.view.layoutIfNeeded() },
                           completion: nil)
        }
    }
    
    @objc func viewRotated()
    {
        DispatchQueue.main.async
            {
                if UIDeviceOrientationIsPortrait(UIDevice.current.orientation)
                {
                    self.routerImage.isHidden = false
                    self.routerLabel.isHidden = false
                }
                
                if UIDeviceOrientationIsLandscape(UIDevice.current.orientation)
                {
                    self.routerImage.isHidden = true
                    self.routerLabel.isHidden = true
                }
        }
    }
    
    @objc func completionTimeout()
    {
        /* Reconnecting */
        referenceController.btManager.connect(btDevice, options: nil)
    }
}
