/*
 * Copyright (c) 2017, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

import GoogleAPIClientForREST
import GTMOAuth2
import GoogleSignIn
import UIKit

class GoogleDriveViewController: UIViewController, GIDSignInUIDelegate, GIDSignInDelegate, UITextFieldDelegate
{
    // MARK: - Properties
    @IBOutlet weak var fileNameField: UITextField!
    @IBOutlet weak var downloadButton: UIButton!
    @IBOutlet weak var logArea: UITextView!
    @IBOutlet weak var logLabel: UILabel!
    @IBOutlet weak var bottomMargin: NSLayoutConstraint!

    // MARK: - Variables
    private let kKeychainItemName = "Drive API"
    private let kClientID = "1024100868464-0iarvpf7nss6movk6moten79i4g5tuqr.apps.googleusercontent.com"
    private let scopes = [kGTLRAuthScopeDriveReadonly]
    private let service = GTLRDriveService()
    var referencedController: FirmwareSourceViewController!
    
    // MARK: - Actions
    @IBAction func downloadButtonPressed(_ sender: Any)
    {
        enableGUIElements(state: false)
        
        if (fileNameField.text?.isEmpty)!
        {
            logArea.text = logArea.text + "\n Filename cannot be empty!"
            enableGUIElements(state: true)
            return
        }
        
        let query = GTLRDriveQuery_FilesList.query()
        query.spaces = "drive"
        query.q = "name = '" + fileNameField.text! + "'"
        
        service.executeQuery(query,
                             delegate: self,
                             didFinish: #selector(GoogleDriveViewController.displayResultWithTicket(ticket:finishedWithObject:error:)))
    }
    
    override func viewDidLoad()
    {
        super.viewDidLoad()
        
        GIDSignIn.sharedInstance().delegate = self
        GIDSignIn.sharedInstance().uiDelegate = self
        GIDSignIn.sharedInstance().scopes = scopes
        fileNameField.delegate = self
        
        /* Adding observer to shift view for keyboard */
        NotificationCenter.default.addObserver(self,
                                               selector: #selector(self.keyboardNotification),
                                               name: NSNotification.Name.UIKeyboardWillChangeFrame,
                                               object: nil)
        
        NotificationCenter.default.addObserver(self,
                                               selector: #selector(GoogleDriveViewController.viewRotated),
                                               name: NSNotification.Name.UIDeviceOrientationDidChange, object: nil)
        
        downloadButton.layer.cornerRadius = 8.0
        downloadButton.clipsToBounds = true
        
        viewRotated()
    }
    
    func completedDownload(ticket: GTLRServiceTicket?,
                          object : Any?,
                           error : Error?) -> Void
    {
        
    }
    
    func enableGUIElements(state: Bool)
    {
        fileNameField.isEnabled = state
        downloadButton.isEnabled = state
    }
    
    override func viewDidAppear(_ animated: Bool)
    {
        GIDSignIn.sharedInstance().signIn()
    }
    
    // Process the response and display output
    @objc func displayResultWithTicket(ticket: GTLRServiceTicket,
                                 finishedWithObject result : GTLRDrive_FileList,
                                 error : NSError?)
    {
        if let error = error
        {
            showAlert(title: "Error", message: error.localizedDescription)
            enableGUIElements(state: true)
            return
        }
        
        if (result.files!.count == 1)
        {
            let foundFile = result.files?[0]
            logArea.text = logArea.text + "\nFile " + (foundFile?.name!)! + " found!"
            let downloadQuery = GTLRDriveQuery_FilesGet.queryForMedia(withFileId: (foundFile?.identifier)!)
            
            service.executeQuery(downloadQuery, completionHandler:
            {
                (ticket: GTLRServiceTicket?, data: Any?, error: Error?) -> Void in
                
                if(error == nil)
                {
                    let file: GTLRDataObject = data as! GTLRDataObject
                    self.logArea.text = self.logArea.text +
                            "\nFile downloaded with size of " + String(file.data.count) + " bytes"
                    
                    let importedImage: ImportedFirmwareImage = ImportedFirmwareImage(firmware: file.data,
                                                                                     source: DownloadSource.FIRMWARE_GOOGLEDRIVE,
                                                                                     filename: self.fileNameField.text!)
                    
                    self.referencedController.referencedController.curFirmware = importedImage
                    self.referencedController.referencedController.firmwareNameLabel.text = importedImage.getTableLabel()
                    self.referencedController.referencedController.firmwareImages.append(importedImage)
                    self.referencedController.referencedController.saveFirmwareImages()
                    
                    self.navigationController!.popToViewController(self.referencedController.referencedController,
                                                                   animated: true)
                }
                else
                {
                    self.logArea.text = self.logArea.text + "\n ERROR! Problem downloading file!"
                    self.enableGUIElements(state: true)
                }
                
            }
            
            )
        }
        else
        {
            logArea.text = logArea.text +  "\nNo files found named " + fileNameField.text!
            enableGUIElements(state: true)
        }
        
    }
    
    // Helper for showing an alert
    func showAlert(title : String, message: String)
    {
        let alert = UIAlertController(
            title: title,
            message: message,
            preferredStyle: UIAlertControllerStyle.alert
        )
        let ok = UIAlertAction(
            title: "OK",
            style: UIAlertActionStyle.default,
            handler: nil
        )
        alert.addAction(ok)
        present(alert, animated: true, completion: nil)
    }

    override func didReceiveMemoryWarning() {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }
    
    // MARK: - GIDSingInUIDelegate
    func sign(inWillDispatch signIn: GIDSignIn!, error: Error!)
    {
        
    }
    
    // Present a view that prompts the user to sign in with Google
    func sign(_ signIn: GIDSignIn!,
              present viewController: UIViewController!)
    {
        self.present(viewController, animated: true, completion: nil)
    }
    
    // Dismiss the "Sign in with Google" view
    func sign(_ signIn: GIDSignIn!,
              dismiss viewController: UIViewController!)
    {
        self.dismiss(animated: true, completion: nil)
    }
    
    func sign(_ signIn: GIDSignIn!, didSignInFor user: GIDGoogleUser!, withError error: Error!)
    {
        if (error == nil)
        {
            service.authorizer = user.authentication.fetcherAuthorizer()
            logArea.text = "Google successfully signed in and authenticated."
            
        }
        else
        {
            self.navigationController!.popToViewController(self.referencedController.referencedController,
                                                           animated: true)
            
        }
    }
    
    func sign(_ signIn: GIDSignIn!, didDisconnectWith user: GIDGoogleUser!, withError error: Error!)
    {
        
    }
    
    @objc func viewRotated()
    {
        DispatchQueue.main.async
            {
                if UIDeviceOrientationIsPortrait(UIDevice.current.orientation)
                {
                    self.logLabel.isHidden = false
                    self.logArea.isHidden = false
                }
                else if UIDeviceOrientationIsLandscape(UIDevice.current.orientation)
                {
                    self.logLabel.isHidden = true
                    self.logArea.isHidden = true
                }
        }
    }
    
    @objc func keyboardNotification(notification: NSNotification)
    {
        
        if let userInfo = notification.userInfo
        {
            let endFrame = (userInfo[UIKeyboardFrameEndUserInfoKey] as? NSValue)?.cgRectValue
            let duration:TimeInterval = (userInfo[UIKeyboardAnimationDurationUserInfoKey] as? NSNumber)?.doubleValue ?? 0
            let animationCurveRawNSN = userInfo[UIKeyboardAnimationCurveUserInfoKey] as? NSNumber
            let animationCurveRaw = animationCurveRawNSN?.uintValue ?? UIViewAnimationOptions.curveEaseInOut.rawValue
            let animationCurve:UIViewAnimationOptions = UIViewAnimationOptions(rawValue: animationCurveRaw)
            if (endFrame?.origin.y)! >= UIScreen.main.bounds.size.height
            {
                bottomMargin?.constant = 10.0
            }
            else
            {
                bottomMargin?.constant = endFrame?.size.height ?? 10.0
            }
            UIView.animate(withDuration: duration,
                           delay: TimeInterval(0),
                           options: animationCurve,
                           animations: { self.view.layoutIfNeeded() },
                           completion: nil)
        }
    }
    
    // MARK: - UITextFieldDelegate
    func textFieldShouldReturn(_ textField: UITextField) -> Bool
    {
        fileNameField.resignFirstResponder()
        return true
    }
    
    func textFieldDidEndEditing(_ textField: UITextField)
    {

    }

}
