/*
 * Copyright (c) 2018, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

import UIKit
import AVFoundation
import LocalAuthentication
import CoreBluetooth

class LockViewController: UIViewController, CBPeripheralDelegate, UITextFieldDelegate
{
    
    // MARK eLock Properties
    let TI_ELOCK_NAME = "MSP432 SAP"
    let TI_ELOCK_SERVICE_UUID = CBUUID(string: "FFF0")
    let TI_ELOCK_CHAR_UUID_WRITE = CBUUID(string: "FFF1")
    let TI_ELOCK_CHAR_UUID_NOTIFY = CBUUID(string: "FFF2")
    var writeArrayLock: [UInt8] = [0x12, 0x34]
    var writeArrayUnlock: [UInt8] = [0x12, 0x34]
    
    // MARK: Variables
    var error: NSError?
    var myLocalizedReasonString = "Authentication is required"
    var lockedState = true
    var buttonEffectPlayer = AVAudioPlayer()
    let lockSoundURL =  Bundle.main.url(forResource: "lock",
                                        withExtension: "wav")!
    let wrongSoundURL =  Bundle.main.url(forResource: "wrong",
                                         withExtension: "mp3")!
    var connectingCon: DiscoveryViewController!
    var referencedController: BLEDeviceTableViewController!
    
    var initializedWrite: Bool = false
    var manualStop: Bool = false
    
    // MARK: Bluetooth parameters
    var btDevice: CBPeripheral!
    var lockCharWrite: CBCharacteristic!
    var lockCharNotify: CBCharacteristic!
    
    // MARK: Properties
    @IBOutlet weak var lockImage: UIImageView!
    @IBOutlet weak var statusLabel: UILabel!
    @IBOutlet weak var keyStack: UIStackView!
    @IBOutlet weak var keyText: UITextField!
    @IBOutlet weak var bottomMargin: NSLayoutConstraint!
    @IBOutlet var doubleTapProp: UITapGestureRecognizer!
    @IBOutlet var singleTapProp: UITapGestureRecognizer!
    
    // MARK: UIViewController
    override func viewDidLoad()
    {
        super.viewDidLoad()
        
        keyText.delegate = self
        singleTapProp.require(toFail: doubleTapProp)
        NotificationCenter.default.addObserver(self,
                                               selector: #selector(self.keyboardNotification),
                                               name: NSNotification.Name.UIKeyboardWillChangeFrame,
                                               object: nil)
        statusLabel.text = "Connected"
        keyText.delegate = self
    }
    
    override func viewWillDisappear(_ animated: Bool)
    {
        referencedController.btManager.scanForPeripherals(withServices: nil, options: nil)
        referencedController.btManager.cancelPeripheralConnection(self.btDevice)
    }
    
    override func didReceiveMemoryWarning()
    {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }
    
    // MARK: Actions
    @IBAction func lockPressedAction(_ sender: AnyObject)
    {
        authenticateUser()
    }
    
    @IBAction func doubleTap(_ sender: UITapGestureRecognizer)
    {
        promptWithKeyPad()
    }
    
    // MARK: Functions
    func lockSuccessful()
    {
        var writeData: Data
        
        print("eLock: Authentication passed!")
        
        if(lockedState)
        {
            writeData = Data(bytes: writeArrayUnlock)
        }
        else
        {
            writeData = Data(bytes: writeArrayLock)
        }
        
        initializedWrite = true
        btDevice.writeValue(writeData, for: lockCharWrite,
                                type: CBCharacteristicWriteType.withResponse)
    }
    
    func lockFailed()
    {
        print("eLock: Authentication failed!")
        
        do
        {
            buttonEffectPlayer = try AVAudioPlayer(contentsOf: wrongSoundURL)
            buttonEffectPlayer.prepareToPlay()
            buttonEffectPlayer.play()
        }
        catch
        {
            print("eLock: ERROR - Could not play effect")
        }
    }
    
    // MARK: Authentication
    func authenticateUser()
    {
        let context = LAContext()
        let reason  = "Authentication for Lock"
        var error: NSError?
        
        if context.canEvaluatePolicy(.deviceOwnerAuthenticationWithBiometrics,
                                     error: &error)
        {
            keyText.resignFirstResponder()
            keyStack.isHidden = true
            
            context.evaluatePolicy(.deviceOwnerAuthenticationWithBiometrics,
                                   localizedReason: reason)
            {
                [unowned self] success, authenticationError in
                
                DispatchQueue.main.async
                    {
                        if success
                        {
                            self.lockSuccessful()
                        }
                        else
                        {
                            self.lockFailed()
                        }
                }
            }
        }
        else
        {
            promptWithKeyPad()
        }
    }
    
    func promptWithKeyPad()
    {
        keyStack.isHidden = false
        keyText.becomeFirstResponder()
    }
    
    // MARK: CBPeripheralDelegate
    func peripheral(_ peripheral: CBPeripheral,
                    didDiscoverServices error: Error?)
    {
        for service in peripheral.services!
        {
            let thisService = service as CBService
            
            if(service.uuid == TI_ELOCK_SERVICE_UUID)
            {
                peripheral.discoverCharacteristics(nil, for: thisService)
                return
            }
        }
    }
    
    /* Callback with characteristic discovery */
    func peripheral(_ peripheral: CBPeripheral,
                    didDiscoverCharacteristicsFor service: CBService,
                    error: Error?)
    {
        for characteristic in service.characteristics!
        {
            let thisCharacteristic = characteristic as CBCharacteristic
            
            if characteristic.uuid == TI_ELOCK_CHAR_UUID_WRITE
            {
                lockCharWrite = thisCharacteristic
            }
            else if characteristic.uuid == TI_ELOCK_CHAR_UUID_NOTIFY
            {
                lockCharNotify = thisCharacteristic
                btDevice.setNotifyValue(true, for: lockCharNotify)
                referencedController.timeoutTimer.invalidate()
                referencedController.connectionCon.dismiss(animated: true, completion: nil)
            }
            else
            {
                print("eLock: Unsupported UUID (\(characteristic.uuid.uuidString))")
            }
        }
    }
    
    /* Callback for a characteristic read/notification */
    func peripheral(_ peripheral: CBPeripheral,
                    didUpdateValueFor characteristic: CBCharacteristic,
                    error: Error?)
    {
        var values = [UInt8](repeating:0, count:characteristic.value!.count)
        characteristic.value?.copyBytes(to: &values,
                                        count: characteristic.value!.count)
        
        /* Play the sound effect */
        do
        {
            buttonEffectPlayer = try AVAudioPlayer(contentsOf: lockSoundURL)
            buttonEffectPlayer.prepareToPlay()
            buttonEffectPlayer.play()
        }
        catch
        {
            print("eLock: ERROR - Could not play effect")
        }
        
        /* If the value is zero, update the UI to unlocked */
        if(values[0] == 0)
        {
            statusLabel.text = "Unlocked"
            lockImage.image = UIImage(named: "unlock")
            lockedState = false
            
        }
            /* Otherwise the value is locked */
        else if(values[0] == 1)
        {
            
            statusLabel.text = "Locked"
            lockImage.image = UIImage(named: "lock")
            lockedState = true
        }
    }
    
    /* Debug log for enabling/disabling notifications */
    func peripheral(_ peripheral: CBPeripheral,
                    didUpdateNotificationStateFor characteristic: CBCharacteristic,
                    error: Error?)
    {
        if(characteristic.isNotifying)
        {
            print("eLock: Notifications enabled for: \(characteristic.uuid.uuidString)")
        }
        else
        {
            print("eLock: Notifications disabled for: \(characteristic.uuid.uuidString)")
        }
    }
    
    /* Call back for input into the keypad. If four characters were written,
     write the value into the BLE lock characteristic */
    @IBAction func charChanged(_ sender: UITextField)
    {
        if sender.text?.characters.count == 4
        {
            keyText.resignFirstResponder()
            keyStack.isHidden = true
            
            let chars = Array(sender.text!.characters)
            
            let numbers = stride(from: 0, to: chars.count, by: 2).map
            {
                UInt8(String(chars[$0 ..< $0+2]), radix: 16) ?? 0
            }
            
            var writeData: Data = Data(bytes: numbers)
            
            if(writeData[0] != 0x12 || writeData[1] != 0x34)
            {
                self.lockFailed()
                return
            }
            else if(lockedState)
            {
                writeData = Data(bytes: writeArrayUnlock)
            }
            else
            {
                writeData = Data(bytes: writeArrayLock)
            }
            
            initializedWrite = true
            btDevice.writeValue(writeData, for: lockCharWrite,
                                    type: CBCharacteristicWriteType.withResponse)
            
        }
    }
    
    /* Clear whatever is currently in the keypad */
    func textFieldDidBeginEditing(_ textField: UITextField)
    {
        keyText.text = ""
    }
    
    /* Adjust GUI items for showing the keyboard */
    func keyboardNotification(notification: NSNotification)
    {
        if let userInfo = notification.userInfo
        {
            let endFrame = (userInfo[UIKeyboardFrameEndUserInfoKey] as? NSValue)?.cgRectValue
            let duration:TimeInterval = (userInfo[UIKeyboardAnimationDurationUserInfoKey] as? NSNumber)?.doubleValue ?? 0
            let animationCurveRawNSN = userInfo[UIKeyboardAnimationCurveUserInfoKey] as? NSNumber
            let animationCurveRaw = animationCurveRawNSN?.uintValue ?? UIViewAnimationOptions.curveEaseInOut.rawValue
            let animationCurve:UIViewAnimationOptions = UIViewAnimationOptions(rawValue: animationCurveRaw)
            if (endFrame?.origin.y)! >= UIScreen.main.bounds.size.height
            {
                bottomMargin?.constant = 15
            }
            else
            {
                bottomMargin?.constant = endFrame?.size.height ?? 60
            }
            UIView.animate(withDuration: duration,
                           delay: TimeInterval(0),
                           options: animationCurve,
                           animations: { self.view.layoutIfNeeded() },
                           completion: nil)
        }
    }
    
}

