/*
 * Copyright (c) 2018, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
import UIKit
import CoreBluetooth
import Charts

class MotorConditionViewController: UIViewController, CBPeripheralDelegate
{
    // MARK: Variables
    let SAP_NOTIFY_UUID = CBUUID(string: "FFF4")
    let SAP_SERVICE_UUID = CBUUID(string: "FFF0")
    var DATA_TO_READ = 0
    var dataRead: Int = 0
    var dataValues = [UInt8]()
    var notifyChar: CBCharacteristic!
    var btDevice: CBPeripheral!
    var btMan: CBCentralManager!
    var lineChartDataSet: LineChartDataSet!
    var lineChartData: LineChartData!
    var lineChartDataSetX: LineChartDataSet!
    var lineChartDataSetY: LineChartDataSet!
    var lineChartDataSetZ: LineChartDataSet!
    var xDataEntries: [ChartDataEntry] = []
    var yDataEntries: [ChartDataEntry] = []
    var zDataEntries: [ChartDataEntry] = []
    var motorData: MotorWearInformation!
    var referenceController: MotorInformationViewController!
    
    // MARK: Properties
    @IBOutlet weak var graphView: LineChartView!
    @IBOutlet weak var titleLabel: UILabel!
    @IBOutlet weak var statusLabel: UILabel!
    @IBOutlet weak var dcStack: UIStackView!
    @IBOutlet weak var coordStack: UIStackView!
    @IBOutlet weak var xSwitch: UISwitch!
    @IBOutlet weak var ySwitch: UISwitch!
    @IBOutlet weak var zSwitch: UISwitch!
    @IBOutlet weak var dcSwitch: UISwitch!
    
    // MARK: Actions
    @IBAction func switchChagned(_ sender: UISwitch)
    {
        updateGraphWithLatestData()
    }
    
    // MARK: Functions
    override func viewDidLoad()
    {
        super.viewDidLoad()

        /* Disabling Screen Timeout */
        UIApplication.shared.isIdleTimerDisabled = true
        
        /* Initializing all of our data sets */
        lineChartDataSetX = LineChartDataSet(values: xDataEntries,
                                             label: "X")
        lineChartDataSetX.setColor(UIColor.red)
        lineChartDataSetY = LineChartDataSet(values: yDataEntries,
                                             label: "Y")
        lineChartDataSetY.setColor(UIColor.blue)
        lineChartDataSetZ = LineChartDataSet(values: zDataEntries,
                                             label: "Z")
        lineChartDataSetZ.setColor(UIColor.green)
        adjustChartFormat()
        
        /* Calculating the number of FFT bytes to read */
        DATA_TO_READ = Int((motorData.xFFTPoints * 4) + (motorData.yFFTPoints * 4) + (motorData.zFFTPoints * 4))
        
        /* Adding observer to rotate view */
        NotificationCenter.default.addObserver(self,
                                               selector: #selector(MotorConditionViewController.viewRotated),
                                               name: NSNotification.Name.UIDeviceOrientationDidChange, object: nil)
        viewRotated()
        btDevice.delegate = self
        
        /* Initiating our BLE connection */
        btMan.connect(btDevice, options: nil)
    }

    override func didReceiveMemoryWarning()
    {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }
    
    /* This function adjusts the format of the chart to something that we like and looks pretty */
    func adjustChartFormat()
    {
        lineChartDataSetX.setCircleColor(UIColor.black)
        lineChartDataSetX.circleRadius = 0.0
        lineChartDataSetX.drawCircleHoleEnabled = false
        lineChartDataSetX.drawValuesEnabled = false
        lineChartDataSetX.lineWidth = 2.0
        
        lineChartDataSetY.setCircleColor(UIColor.black)
        lineChartDataSetY.circleRadius = 0.0
        lineChartDataSetY.drawCircleHoleEnabled = false
        lineChartDataSetY.drawValuesEnabled = false
        lineChartDataSetY.lineWidth = 2.0
        
        lineChartDataSetZ.setCircleColor(UIColor.black)
        lineChartDataSetZ.circleRadius = 0.0
        lineChartDataSetZ.drawCircleHoleEnabled = false
        lineChartDataSetZ.drawValuesEnabled = false
        lineChartDataSetZ.lineWidth = 2.0
        
        lineChartData = LineChartData()
        lineChartData.addDataSet(lineChartDataSetX)
        lineChartData.addDataSet(lineChartDataSetY)
        lineChartData.addDataSet(lineChartDataSetZ)
        
        graphView.autoScaleMinMaxEnabled = true
        
        graphView.chartDescription?.text = "Frequency Distribution"
        graphView.xAxis.labelPosition = XAxis.LabelPosition.bottom
        graphView.data = lineChartData
    }

    // MARK: CBPeripheralDelegate
    func peripheral(_ peripheral: CBPeripheral,
                    didDiscoverServices error: Error?)
    {
        for service in peripheral.services!
        {
            if service.uuid == SAP_SERVICE_UUID
            {
                let thisService = service as CBService
                peripheral.discoverCharacteristics(nil, for: thisService)
                return
            }
        }
    }
    
    /* This gets called when we get a characteristic read. When this happens, if the notify characeristic if found
        we start to initiate the transfer of FFT data via the notification */
    func peripheral(_ peripheral: CBPeripheral,
                    didDiscoverCharacteristicsFor service: CBService,
                    error: Error?)
    {
        for characteristic in service.characteristics!
        {
            if characteristic.uuid == SAP_NOTIFY_UUID
            {
                notifyChar = characteristic
            }
        }
        
        if(notifyChar != nil)
        {
            dataRead = 0
            dataValues = [UInt8](repeating:0, count: DATA_TO_READ)
            btDevice.setNotifyValue(true, for: notifyChar)
        }
    }
    
    /* Callback for a characteristic read/notification */
    func peripheral(_ peripheral: CBPeripheral,
                    didUpdateValueFor characteristic: CBCharacteristic,
                    error: Error?)
    {
        if characteristic.uuid == SAP_NOTIFY_UUID
        {
            characteristic.value?.copyBytes(to: &dataValues + dataRead,
                                            count: characteristic.value!.count)
            dataRead = dataRead + characteristic.value!.count
            
            /* If we read all of the data we need to update the graph */
            if(dataRead == DATA_TO_READ)
            {
   
                btMan.cancelPeripheralConnection(btDevice)
                updateGraphWithLatestData()
                referenceController.connectionCon.dismiss(animated: true, completion: nil)
                referenceController.timeoutTimer.invalidate()
                statusLabel.text =  "Data Received"
            }
        }
        else
        {
            print("SAP: Invalid characteristic read (\(characteristic.uuid.uuidString))")
            return
        }
    }
    
    override func viewWillDisappear(_ animated: Bool)
    {
        btMan.cancelPeripheralConnection(self.btDevice)
    }
    
    /* Takes the data received and updates the graph. This function will account for the various pickers
        on the GUI */
    func updateGraphWithLatestData()
    {
        var xIncrement: Float = (Float(motorData.xSamplingRate)/2.0)
        xIncrement = xIncrement / Float(motorData.xFFTPoints)
        var curX: Float = 0.0
        
        lineChartDataSetX.clear()
        lineChartDataSetY.clear()
        lineChartDataSetZ.clear()
        
        var yOffset: Int = Int(0)
        var zOffset: Int = Int(0)
        
        for index in stride(from: 0, to:Int(motorData.xFFTPoints*4), by: 4)
        {
            var floatValue: Float32 = 0
            var tempArray = [UInt8]()
            
            /* Calculating the offsets */
            yOffset = index + Int(motorData.xFFTPoints*4)
            zOffset = index + Int(motorData.xFFTPoints*8)
            
            /* Skip this if DC is hidden */
            if((index == 0) && (dcSwitch.isOn == false))
            {
                curX = curX + xIncrement
                continue
            }
            
            /* X Value */
            if(xSwitch.isOn == true)
            {
                tempArray = Array(dataValues[index...(index+3)])
                memcpy(&floatValue, tempArray, 4)
                _ = lineChartDataSetX.addEntry(ChartDataEntry(x: Double(curX),
                                                              y: Double(floatValue)))
            }
            
            /* Y Value */
            if(ySwitch.isOn == true)
            {
                tempArray = Array(dataValues[(yOffset)...(yOffset+3)])
                memcpy(&floatValue, tempArray, 4)
                _ = lineChartDataSetY.addEntry(ChartDataEntry(x: Double(curX),
                                                              y: Double(floatValue)))
            }
            
            /* Z Value */
            if(zSwitch.isOn)
            {
                tempArray = Array(dataValues[(zOffset)...(zOffset+3)])
                memcpy(&floatValue, tempArray, 4)
                _ = lineChartDataSetZ.addEntry(ChartDataEntry(x: Double(curX),
                                                              y: Double(floatValue)))
            }
            
            curX = curX + xIncrement
        }

        /* Notify and update the GUI */
        lineChartData.notifyDataChanged()
        graphView.notifyDataSetChanged()
    }
    
    @objc func viewRotated()
    {
        DispatchQueue.main.async
            {
                if UIDeviceOrientationIsPortrait(UIDevice.current.orientation)
                {
                    self.statusLabel.isHidden = false
                    self.titleLabel.isHidden = false
                    self.dcStack.isHidden = false
                    self.coordStack.isHidden = false
                }
                else if UIDeviceOrientationIsLandscape(UIDevice.current.orientation)
                {
                    self.statusLabel.isHidden = true
                    self.titleLabel.isHidden = true
                    self.dcStack.isHidden = true
                    self.coordStack.isHidden = true
                }
        }
    }

}
