/*
 * Copyright (c) 2018, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
import UIKit
import CoreBluetooth

class MotorDeviceViewController: UIViewController, UITableViewDelegate, UITableViewDataSource,
    CBCentralManagerDelegate
{
    // MARK: Properties:
    @IBOutlet var deviceTable: UITableView!
    @IBOutlet weak var dirLabel: UILabel!
    
    // MARK: Variables
    var btManager: CBCentralManager!
    var btDevices: [CBPeripheral]! = []
    var selBTDev: CBPeripheral!
    var rssiReadings: [NSNumber]! = []
    var advData: [[UInt8]]! = []
    let identifier = "MotorDeviceViewCell"
    var codeExampleTable: UIViewController!
    var advertisementData: [UInt8]!
    
    override func viewDidLoad()
    {
        super.viewDidLoad()
        btManager = CBCentralManager(delegate: self, queue: nil)
        deviceTable.dataSource = self
        deviceTable.delegate = self
        dirLabel.adjustsFontSizeToFitWidth = true
    }
    
    override func didReceiveMemoryWarning()
    {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }
    
    // MARK: - Table view data source
    func numberOfSections(in tableView: UITableView) -> Int
    {
        return 1
    }
    
    func tableView(_ tableView: UITableView,
                            numberOfRowsInSection section: Int) -> Int
    {
        return btDevices.count
    }
    
    func tableView(_ tableView: UITableView, heightForRowAt indexPath: IndexPath) -> CGFloat
    {
        return 40.0
    }
    
    func tableView(_ tableView: UITableView,
                            cellForRowAt indexPath: IndexPath) -> UITableViewCell
    {
        let cell: MotorDeviceTableViewCell = tableView.dequeueReusableCell(withIdentifier: identifier, for:
            indexPath) as! MotorDeviceTableViewCell
        let device = btDevices[indexPath.row]
        let devName = device.name
        
        if devName != nil
        {
            cell.deviceText.text = device.name
        }
        else
        {
            cell.deviceText.text = "Unnamed Device"
        }
        
        cell.setSignalStrengthImage(rssi: rssiReadings[indexPath.row].floatValue)
        
        return cell
    }
    
    func tableView(_ tableView: UITableView,
                            didSelectRowAt indexPath: IndexPath)
    {
        advertisementData = advData[indexPath.row]
        selBTDev = btDevices[indexPath.row]
        self.performSegue(withIdentifier: "motorSegue", sender: self)
    }
    
    // MARK: CBCentralManagerDelegate
    func centralManagerDidUpdateState(_ central: CBCentralManager)
    {
        if central.state == .poweredOn
        {
            central.scanForPeripherals(withServices: nil, options: nil)
        }
        else
        {
            _ = self.navigationController?.popViewController(animated: true);
        }
    }
    
    func centralManager(_ central: CBCentralManager,
                        didConnect peripheral: CBPeripheral)
    {
        peripheral.discoverServices(nil)
    }
    
    func centralManager(_ central: CBCentralManager,
                        didDiscover peripheral: CBPeripheral,
                        advertisementData: [String : Any],
                        rssi RSSI: NSNumber)
    {
        if(!btDevices.contains(peripheral))
        {
            var insertRow = 0
            
            while(insertRow < rssiReadings.count)
            {
                if(rssiReadings[insertRow].doubleValue < RSSI.doubleValue)
                {
                    break
                }
                
                insertRow = insertRow + 1
            }
            
            let manData = advertisementData[CBAdvertisementDataManufacturerDataKey] as? Data
            
            /* Making sure that we actually have manufacturer data */
            if(manData == nil)
            {
                return
            }
            
            let byteArray = [UInt8](manData!)
            
            if(byteArray.count == 26 && byteArray[0] == 0x54 && byteArray[1] == 0x49)
            {
                let newIndexPath = IndexPath(row: insertRow, section: 0)
                btDevices.insert(peripheral, at: insertRow)
                rssiReadings.insert(RSSI, at: insertRow)
                deviceTable.insertRows(at: [newIndexPath], with: .bottom)
                advData.append(byteArray)
            }
        }
    }
    
    func centralManager(_ central: CBCentralManager,
                        didDisconnectPeripheral peripheral: CBPeripheral,
                        error: Error?)
    {
        
    }
    
    @objc func completionTimeout()
    {
        btManager.cancelPeripheralConnection(selBTDev)
        
        let alert = UIAlertController(title: "Connection Timeout",
                                      message: "Timeout occurred while connecting to device. Please verify correct" +
            " services are being advertised on device.",
                                      preferredStyle: UIAlertControllerStyle.alert)
        alert.addAction(UIAlertAction(title: "OK",
                                      style: UIAlertActionStyle.default,
                                      handler: nil))
        _ = navigationController?.popToViewController(self, animated: true)
        present(alert, animated: true, completion: nil)
    }
    
    // MARK: - Navigation
    override func prepare(for segue: UIStoryboardSegue, sender: Any?)
    {
        let motorEnergyCon: MotorInformationViewController = segue.destination as! MotorInformationViewController
        motorEnergyCon.btDevice = selBTDev
        motorEnergyCon.referencedController  = self
        motorEnergyCon.advData = advertisementData
        motorEnergyCon.btMan = btManager
        
        btManager.stopScan()
        btDevices.removeAll()
        rssiReadings.removeAll()
        advData.removeAll()
        deviceTable.reloadData()
    }
    
    override func viewWillAppear(_ animated: Bool)
    {
        if(btManager != nil)
        {
            btManager.scanForPeripherals(withServices: nil, options: nil)
        }
    }
}
