/*
 * Copyright (c) 2018, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
import UIKit
import CoreBluetooth

// MARK: Structure Definition
struct MotorWearInformation
{
    var xFFTPoints: Float32 = 0.0
    var xSamplingRate: UInt32 = 0
    var xMagnitude: Float32 = 0.0
    var xFrequencyMax: Float32 = 0.0
    var yFFTPoints: Float32 = 0.0
    var ySamplingRate: UInt32 = 0
    var yMagnitude: Float32 = 0.0
    var yFrequencyMax: Float32 = 0.0
    var zFFTPoints: Float32 = 0.0
    var zSamplingRate: UInt32 = 0
    var zMagnitude: Float32 = 0.0
    var zFrequencyMax: Float32 = 0.0
}

class MotorInformationViewController: UIViewController
{
    // MARK: Properties
    @IBOutlet weak var connectButton: UIButton!
    @IBOutlet weak var energyLabel: UILabel!
    @IBOutlet weak var xFreq: UILabel!
    @IBOutlet weak var xMag: UILabel!
    @IBOutlet weak var yFreq: UILabel!
    @IBOutlet weak var yMag: UILabel!
    @IBOutlet weak var zFreq: UILabel!
    @IBOutlet weak var zMag: UILabel!
    @IBOutlet weak var gearImage: UIImageView!

    // MARK: Variables
    var btDevice: CBPeripheral!
    var btMan: CBCentralManager!
    var referencedController: UIViewController!
    var advData: [UInt8]!
    var motorData: MotorWearInformation!
    var connectionCon: DiscoveryViewController!
    var timeoutTimer: Timer!
    
    // MARK: Actions
    @IBAction func connectPressed(_ sender: Any)
    {
        connectionCon = DiscoveryViewController(message: "Connecting...")
        present(connectionCon, animated: true, completion: nil)
        
        timeoutTimer = Timer.scheduledTimer(timeInterval: 15.0,
                                            target: self,
                                            selector: #selector(self.completionTimeout),
                                            userInfo: nil,
                                            repeats: false)
    }
    
    // MARK: Functions
    override func viewDidLoad()
    {
        super.viewDidLoad()
        
        /* Adjusting the GUI Components */
        energyLabel.clipsToBounds = true
        connectButton.clipsToBounds = true
        energyLabel.layer.cornerRadius = 8.0
        connectButton.layer.cornerRadius = 8.0
        
        /* Calculating the maximum energies */
        motorData = MotorWearInformation()
        
        /* X Values */
        var freqMagnitude: Float32 = 0.0
        var numOfFFTPoints = pow(2.0,(Float(advData[2] >> 5) + 7))
        var samplingFreq: UInt32 = UInt32((advData[4]))
        samplingFreq = (samplingFreq << 8) | UInt32(advData[5])
        var binIndex: UInt32 = UInt32(((advData[2] & 0x1F) << 13) | advData[3])
        var calcFreq: Float = Float((binIndex * samplingFreq)) / Float(numOfFFTPoints * 2)
        var tempArray: [UInt8] = Array(advData[6...9])
        memcpy(&freqMagnitude, tempArray, 4)
        motorData.xFFTPoints = numOfFFTPoints
        motorData.xFrequencyMax = calcFreq
        motorData.xMagnitude = freqMagnitude
        motorData.xSamplingRate = samplingFreq
        
        xMag.text = "Magnitude: " + String.localizedStringWithFormat("%.2f", freqMagnitude)
        xFreq.text = "Frequency: " + String.localizedStringWithFormat("%.2fHz", calcFreq)
        
        /* Y Values */
        numOfFFTPoints = pow(2.0,(Float(advData[10] >> 5) + 7))
        samplingFreq = UInt32((advData[12]))
        samplingFreq = (samplingFreq << 8) | UInt32(advData[13])
        binIndex = UInt32(((advData[10] & 0x1F) << 13) | advData[11])
        calcFreq = Float((binIndex * samplingFreq)) / Float(numOfFFTPoints * 2)
        tempArray = Array(advData[14...17])
        memcpy(&freqMagnitude, tempArray, 4)
        motorData.yFFTPoints = numOfFFTPoints
        motorData.yFrequencyMax = calcFreq
        motorData.yMagnitude = freqMagnitude
        motorData.ySamplingRate = samplingFreq
        
        yMag.text = "Magnitude: " + String.localizedStringWithFormat("%.2f", freqMagnitude)
        yFreq.text = "Frequency: " + String.localizedStringWithFormat("%.2fHz", calcFreq)
        
        /* Z Values */
        numOfFFTPoints = pow(2.0,(Float(advData[18] >> 5) + 7))
        samplingFreq = UInt32((advData[20]))
        samplingFreq = (samplingFreq << 8) | UInt32(advData[21])
        binIndex = UInt32(((advData[18] & 0x1F) << 13) | advData[19])
        calcFreq = Float((binIndex * samplingFreq)) / Float(numOfFFTPoints * 2)
        tempArray = Array(advData[22...25])
        memcpy(&freqMagnitude, tempArray, 4)
        motorData.zFFTPoints = numOfFFTPoints
        motorData.zFrequencyMax = calcFreq
        motorData.zMagnitude = freqMagnitude
        motorData.zSamplingRate = samplingFreq
        
        zMag.text = "Magnitude: " + String.localizedStringWithFormat("%.2f", freqMagnitude)
        zFreq.text = "Frequency: " + String.localizedStringWithFormat("%.2fHz", calcFreq)
        
        NotificationCenter.default.addObserver(self,
                                               selector: #selector(MotorInformationViewController.viewRotated),
                                               name: NSNotification.Name.UIDeviceOrientationDidChange, object: nil)

        viewRotated()
        
    }

    override func didReceiveMemoryWarning()
    {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }
    
    override func prepare(for segue: UIStoryboardSegue, sender: Any?)
    {
        let motorCon: MotorConditionViewController =
            segue.destination as! MotorConditionViewController
        
        motorCon.btDevice = self.btDevice
        motorCon.btMan = self.btMan
        motorCon.motorData = self.motorData
        motorCon.referenceController = self
    }
    
    @objc func completionTimeout()
    {
        btMan.cancelPeripheralConnection(btDevice)
        connectionCon.dismiss(animated: true, completion: nil)
        
        let alert = UIAlertController(title: "Connection Timeout",
                                      message: "Timeout occurred while connecting to device. Please verify correct" +
            " services are being advertised on device.",
                                      preferredStyle: UIAlertControllerStyle.alert)
        alert.addAction(UIAlertAction(title: "OK",
                                      style: UIAlertActionStyle.default,
                                      handler: nil))
        _ = navigationController?.popToViewController(self, animated: true)
        present(alert, animated: true, completion: nil)
    }
    
    @objc func viewRotated()
    {
        DispatchQueue.main.async
        {
            if UIDeviceOrientationIsPortrait(UIDevice.current.orientation)
            {
                self.gearImage.isHidden = false
                self.energyLabel.isHidden = false
            }
            
            if UIDeviceOrientationIsLandscape(UIDevice.current.orientation)
            {
                self.gearImage.isHidden = true
                self.energyLabel.isHidden = true
            }
        }
    }
}
