import UIKit
import Charts
import CoreBluetooth

class OpticalViewController: UIViewController
{
    // MARK: Variables
    var referencedController: SensorTabBarController!
    var lineChartDataSet: LineChartDataSet!
    var lineChartData: LineChartData!
    var optDataEntries: [ChartDataEntry] = []
    var curIndex = 0
    
    // MARK: Properties
    @IBOutlet weak var graphView: LineChartView!
    @IBOutlet weak var avgLabel: UILabel!
    @IBOutlet weak var maxLabel: UILabel!
    @IBOutlet weak var minLabel: UILabel!
    @IBOutlet weak var periodLabel: UILabel!
    @IBOutlet weak var periodSlider: UISlider!
    @IBOutlet weak var enableSwitch: UISwitch!
    @IBOutlet weak var valLabel: UILabel!
    @IBOutlet weak var statsView: UIView!
    @IBOutlet weak var optionsStack: UIStackView!
    
    // MARK: Actions
    @IBAction func periodChanged(_ sender: UISlider)
    {
        var dataArray: [UInt8]
        
        periodLabel.text = String(format: "%.1f", sender.value) + "s"
        let setValue = UInt16(sender.value * 1000.0)
        
        dataArray = [UInt8(setValue & 0xFF), UInt8(setValue >> 8)]
        let writeArray = Data(dataArray)
        
        referencedController.btDevice.writeValue(writeArray, for: referencedController.optPeriChar,
                                                 type: CBCharacteristicWriteType.withResponse)
    }
    
    @IBAction func enableChanged(_ sender: UISwitch)
    {
        var dataArray: [UInt8]
        
        if(sender.isOn == true)
        {
            dataArray = [0x01]
        }
        else
        {
            dataArray = [0x00]
        }
        
        let writeArray = Data(dataArray)
        
        referencedController.btDevice.writeValue(writeArray, for: referencedController.optConfChar,
                                                 type: CBCharacteristicWriteType.withResponse)
    }
    
    // MARK: Functions
    func updateStatistics(stats: SensorStatistics)
    {
        maxLabel.text = String(format: "%.1f", stats.max)
        minLabel.text = String(format: "%.1f", stats.min)
        avgLabel.text = String(format: "%.1f", stats.avg)
    }
    
    func adjustChartFormat()
    {
        let gradientColors = [UIColor(red: 0.5, green: 0.0, blue: 0.0, alpha: 0.7).cgColor,
                              UIColor(red: 1.0, green: 0.0, blue: 0.0, alpha: 0.7).cgColor]
        let colorLocations: [CGFloat] = [1.0, 0.0]
        let gradient = CGGradient(colorsSpace: CGColorSpaceCreateDeviceRGB(),
                                  colors: gradientColors as CFArray,
                                  locations: colorLocations)
        lineChartDataSet.fill = Fill.fillWithLinearGradient(gradient!,
                                                            angle: 90.0)
        lineChartDataSet.setCircleColor(UIColor.black)
        lineChartDataSet.circleRadius = 1.0
        lineChartDataSet.drawCircleHoleEnabled = false
        lineChartDataSet.fillAlpha = 1.0
        lineChartDataSet.drawFilledEnabled = true
        lineChartDataSet.drawValuesEnabled = false
        lineChartData = LineChartData()
        lineChartData.addDataSet(lineChartDataSet)
        graphView.data = lineChartData
        
        /* Adjusting the Margins */
        let yAxis = graphView.leftAxis
        
        yAxis.axisMaximum = 7000
        yAxis.axisMinimum = 0
        
        let xAxis = graphView.xAxis
        xAxis.enabled = false
        
        graphView.chartDescription?.text = "Optical Data"
        
    }
    
    func addGraphValue(floatValue: Float)
    {
        if (curIndex == 0)
        {
            lineChartDataSet = LineChartDataSet(values: optDataEntries,
                                                label: "Optical Data")
            adjustChartFormat()
        }
        
        let dataEntry = ChartDataEntry(x: Double(curIndex),
                                       y: Double(floatValue))
        
        if(curIndex == SENSOR_GRAPH_SAMPLES)
        {
            _ = lineChartDataSet.removeFirst()
            
            for dataValue in lineChartDataSet.values
            {
                dataValue.x = dataValue.x - 1
            }
            
        }
        else
        {
            curIndex = curIndex + 1
        }
        
        _ = lineChartDataSet.addEntry(dataEntry)
        lineChartData.notifyDataChanged()
        graphView.notifyDataSetChanged()
        valLabel.text = String(format: "%.1f", floatValue) + " lux"
        
    }
    
    @objc func viewRotated()
    {
        DispatchQueue.main.async
            {
                if UIDeviceOrientationIsPortrait(UIDevice.current.orientation)
                {
                    self.optionsStack.isHidden = false
                    self.statsView.isHidden = false
                }
                else if UIDeviceOrientationIsLandscape(UIDevice.current.orientation)
                {
                    self.optionsStack.isHidden = true
                    self.statsView.isHidden = true
                }
        }
    }
    
    override func viewDidLoad()
    {
        super.viewDidLoad()
        referencedController = tabBarController as! SensorTabBarController
        referencedController.opticalGraphController = self
        
        /* Adding observer to rotate view */
        NotificationCenter.default.addObserver(self,
                                               selector: #selector(OpticalViewController.viewRotated),
                                               name: NSNotification.Name.UIDeviceOrientationDidChange, object: nil)
        viewRotated()
    }
    
    override func didReceiveMemoryWarning() {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }
    
    
}
