/*
 * Copyright (c) 2017, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

import UIKit
import SystemConfiguration.CaptiveNetwork

class RouterSetupViewController: UIViewController, UITextFieldDelegate, UIPickerViewDelegate, UIPickerViewDataSource
{
    // MARK: Variables
    var curSettings = WiFiSettings()
    
    let securityTypes =
    [
        WiFiSecurityType.None,
        WiFiSecurityType.WPA,
        WiFiSecurityType.WEP
    ]
    
    // MARK: Properties
    @IBOutlet weak var ssidField: UITextField!
    @IBOutlet weak var keyField: UITextField!
    @IBOutlet weak var nextButton: UIButton!
    @IBOutlet weak var bottomMargin: NSLayoutConstraint!
    @IBOutlet weak var secTypePicker: UIPickerView!
    
    // MARK: Actions
    @IBAction func nextButtonPressed(_ sender: Any)
    {
        ssidField.resignFirstResponder()
        keyField.resignFirstResponder()
        
        if(ssidField.text?.isEmpty)!
        {
            let alert = UIAlertController(title: "SSID is blank",
                                          message: "Value must be specified for SSID of SimpleLink device!",
                                          preferredStyle: UIAlertControllerStyle.alert)
            alert.addAction(UIAlertAction(title: "OK", style: UIAlertActionStyle.default, handler: nil))
            self.present(alert, animated: true, completion: nil)
            return
        }
        if(secTypePicker.selectedRow(inComponent: 0) != 0 && (keyField.text?.isEmpty)!)
        {
            let alert = UIAlertController(title: "Security Key is Blank",
                                          message: "If security key is required it must not be and empty string!",
                                          preferredStyle: UIAlertControllerStyle.alert)
            alert.addAction(UIAlertAction(title: "OK", style: UIAlertActionStyle.default, handler: nil))
            self.present(alert, animated: true, completion: nil)
            return
        }
        
        if(secTypePicker.selectedRow(inComponent: 0) != 0)
        {
            curSettings.routerKey = keyField.text!
            curSettings.routerSecType = securityTypes[secTypePicker.selectedRow(inComponent: 0)]
        }
        
        curSettings.routerName = ssidField.text!

        
        self.performSegue(withIdentifier: "provisionSummarySegue", sender: self)
    }
    
    @IBAction func nextBarItemPressed(_ sender: Any)
    {
        self.nextButtonPressed(nextButton)
    }
    
    @IBAction func cancelBarItemPressed(_ sender: Any)
    {
        self.navigationController?.popViewController(animated: true)
    }
    
    // MARK: - UITextFieldDelegate
    func textFieldShouldReturn(_ textField: UITextField) -> Bool
    {
        if(textField == ssidField)
        {
            ssidField.resignFirstResponder()
        }
        else
        {
            keyField.resignFirstResponder()
        }
        
        return true
    }
    
    func textFieldDidEndEditing(_ textField: UITextField)
    {
        
    }
    
    // MARK: Functions
    @objc func keyboardNotification(notification: NSNotification)
    {
        
        if let userInfo = notification.userInfo
        {
            let endFrame = (userInfo[UIKeyboardFrameEndUserInfoKey] as? NSValue)?.cgRectValue
            let duration:TimeInterval = (userInfo[UIKeyboardAnimationDurationUserInfoKey] as? NSNumber)?.doubleValue ?? 0
            let animationCurveRawNSN = userInfo[UIKeyboardAnimationCurveUserInfoKey] as? NSNumber
            let animationCurveRaw = animationCurveRawNSN?.uintValue ?? UIViewAnimationOptions.curveEaseInOut.rawValue
            let animationCurve:UIViewAnimationOptions = UIViewAnimationOptions(rawValue: animationCurveRaw)
            if (endFrame?.origin.y)! >= UIScreen.main.bounds.size.height
            {
                bottomMargin?.constant = 10.0
            }
            else
            {
                bottomMargin?.constant = endFrame?.size.height ?? 10.0
            }
            UIView.animate(withDuration: duration,
                           delay: TimeInterval(0),
                           options: animationCurve,
                           animations: { self.view.layoutIfNeeded() },
                           completion: nil)
        }
    }
    
    func currentSSIDs() -> [String]
    {
        guard let interfaceNames = CNCopySupportedInterfaces() as? [String] else
        {
            return []
        }
        return interfaceNames.flatMap { name in
            guard let info = CNCopyCurrentNetworkInfo(name as CFString) as? [String:AnyObject] else
            {
                return nil
            }
            guard let ssid = info[kCNNetworkInfoKeySSID as String] as? String else
            {
                return nil
            }
            return ssid
        }
    }
    
    override func prepare(for segue: UIStoryboardSegue, sender: Any?)
    {
        if(segue.identifier == "provisionSummarySegue")
        {
            let provSummary: ProvisionSummaryViewController =
                segue.destination as! ProvisionSummaryViewController
            provSummary.curSettings = self.curSettings
        }
    }
    
    // MARK: UIPickerViewDataSource
    func numberOfComponents(in pickerView: UIPickerView) -> Int
    {
        return 1
    }
    
    func pickerView(_ pickerView: UIPickerView, numberOfRowsInComponent component: Int) -> Int
    {
        return securityTypes.count
    }
    
    func pickerView(_ pickerView: UIPickerView, titleForRow row: Int, forComponent component: Int) -> String?
    {
        return securityTypes[row].rawValue
    }
    
    func pickerView(_ pickerView: UIPickerView, didSelectRow row: Int, inComponent component: Int)
    {
        if(securityTypes[row] == WiFiSecurityType.None)
        {
            keyField.isEnabled = false
        }
        else
        {
            keyField.isEnabled = true
        }
    }
    
    override func viewDidLoad()
    {
        super.viewDidLoad()

        /* Getting list of connected SSIDs so we can prepopulate */
        let currentNet = currentSSIDs()
        
        if(currentNet.count > 0)
        {
            ssidField.text = currentNet[0]
        }
        
        /* Setting Delegates */
        ssidField.delegate = self
        keyField.delegate = self
        secTypePicker.dataSource = self
        secTypePicker.delegate = self
        
        /* Customizing the GUI */
        nextButton.clipsToBounds = true
        nextButton.layer.cornerRadius = 8.0
        
        /* Adding observer to shift view for keyboard */
        NotificationCenter.default.addObserver(self,
                                               selector: #selector(self.keyboardNotification),
                                               name: NSNotification.Name.UIKeyboardWillChangeFrame,
                                               object: nil)
        
        /* Adding observer to rotate view */
        NotificationCenter.default.addObserver(self,
                                               selector: #selector(RouterSetupViewController.viewRotated),
                                               name: NSNotification.Name.UIDeviceOrientationDidChange, object: nil)
    }
    
    override func viewWillAppear(_ animated: Bool)
    {
  
    }
    
    override func viewWillDisappear(_ animated: Bool)
    {
        /* Removin observer to center the text */
        super.viewWillDisappear(animated)
    }
    
    override func observeValue(forKeyPath keyPath: String?, of object: Any?, change: [NSKeyValueChangeKey : Any]?,
                               context: UnsafeMutableRawPointer?)
    {
        let textView = object as! UITextView
        var topCorrect = (textView.bounds.size.height - textView.contentSize.height * textView.zoomScale) / 2
        topCorrect = topCorrect < 0.0 ? 0.0 : topCorrect;
        textView.contentInset.top = topCorrect
    }

    override func didReceiveMemoryWarning()
    {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }
    
    @objc func viewRotated()
    {
        DispatchQueue.main.async
            {
                if UIDeviceOrientationIsPortrait(UIDevice.current.orientation)
                {
                    self.nextButton.isHidden = false
                }
                
                if UIDeviceOrientationIsLandscape(UIDevice.current.orientation)
                {
                    self.nextButton.isHidden = true
                }
        }
    }
}
