/*
 * Copyright (c) 2017, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

import UIKit
import CoreBluetooth

class SimpleAPViewController: UIViewController, CBPeripheralDelegate, UITextFieldDelegate
{
    // MARK: Variables
    let SAP_SERVICE_UUID = CBUUID(string: "FFF0")
    let SAP_LED_UUID = CBUUID(string: "FFF1")
    let SAP_WRITE_UUID = CBUUID(string: "FFF3")
    let SAP_READ_UUID = CBUUID(string: "FFF2")
    let SAP_NOTIFY_UUID = CBUUID(string: "FFF4")
    var ledChar: CBCharacteristic!
    var writeChar: CBCharacteristic!
    var readChar: CBCharacteristic!
    var notifyChar: CBCharacteristic!
    var btDevice: CBPeripheral!
    var referencedController: BLEDeviceTableViewController!
    var isCurrentlyNotifying: Bool = false
    let hexSet = NSCharacterSet(charactersIn: "0123456789ABCDEFabcdef").inverted
    
    // MARK: Properties
    @IBOutlet weak var ledView: UIView!
    @IBOutlet weak var ledLabel: UILabel!
    @IBOutlet weak var ledButton: UIButton!
    @IBOutlet weak var readView: UIView!
    @IBOutlet weak var readLabel: UILabel!
    @IBOutlet weak var readValue: UILabel!
    @IBOutlet weak var readButton: UIButton!
    @IBOutlet weak var writeView: UIView!
    @IBOutlet weak var writeLabel: UILabel!
    @IBOutlet weak var writeButton: UIButton!
    @IBOutlet weak var notifyView: UIView!
    @IBOutlet weak var notifyLabel: UILabel!
    @IBOutlet weak var notifyValue: UILabel!
    @IBOutlet weak var notifyButton: UIButton!
    @IBOutlet weak var readStatus: UILabel!
    @IBOutlet weak var writeStatus: UILabel!
    @IBOutlet weak var notifyStatus: UILabel!
    @IBOutlet weak var writeField: UITextField!
    @IBOutlet weak var bottomMargin: NSLayoutConstraint!
    
    
    // MARK: Actions
    @IBAction func toggleLEDPressed(_ sender: Any)
    {
        let dataArray: [UInt8] = [0x01]
        let writeArray = Data(dataArray)
        btDevice.writeValue(writeArray, for: ledChar,
                            type: CBCharacteristicWriteType.withResponse)
    }
    
    @IBAction func readValuePressed(_ sender: Any)
    {
        btDevice.readValue(for: readChar)
    }
    
    @IBAction func writeButtonPressed(_ sender: Any)
    {
        if(writeField.text == nil || (writeField.text?.isEmpty)!)
        {
            writeStatus.text = "Empty Value!"
            return
        }
        
        let dataArr: [UInt8] = Array(writeField.text!.utf8)
        let writeArray = Data(dataArr)
        btDevice.writeValue(writeArray, for: writeChar, type: CBCharacteristicWriteType.withResponse)
        writeStatus.text = "Value Written"
    }

    
    @IBAction func notifyPressed(_ sender: Any)
    {
        if(self.isCurrentlyNotifying)
        {
            btDevice.setNotifyValue(false, for: notifyChar)
            self.notifyStatus.text = "Disabled"
            self.notifyButton.setTitle("Enable Notifications", for: .normal)
            self.isCurrentlyNotifying = false
        }
        else
        {
            btDevice.setNotifyValue(true, for: notifyChar)
            self.notifyStatus.text = "Enabled"
            self.notifyButton.setTitle("Disable Notifications", for: .normal)
            self.isCurrentlyNotifying = true
        }
    }
    
    // MARK: Functions
    override func viewDidLoad()
    {
        super.viewDidLoad()
        
        /* Tweaking the GUI a bit */
        readLabel.clipsToBounds = true
        readButton.layer.cornerRadius = 8.0
        readLabel.layer.cornerRadius = 8.0
        readLabel.layer.borderColor = UIColor.black.cgColor
        readLabel.layer.borderWidth = 1.0
        writeLabel.layer.cornerRadius = 8.0
        writeButton.layer.cornerRadius = 8.0
        ledButton.layer.cornerRadius = 8.0
        ledLabel.layer.cornerRadius = 8.0
        ledLabel.clipsToBounds = true
        ledLabel.layer.borderWidth = 1.0
        ledLabel.layer.borderColor = UIColor.black.cgColor
        ledButton.layer.borderColor = UIColor.black.cgColor
        ledButton.layer.borderWidth = 1.0
        readButton.layer.borderWidth = 1.0
        readButton.layer.borderColor = UIColor.black.cgColor
        writeButton.layer.borderWidth = 1.0
        writeButton.layer.borderColor = UIColor.black.cgColor
        writeLabel.clipsToBounds = true
        writeLabel.layer.borderWidth = 1.0
        notifyLabel.clipsToBounds = true
        notifyLabel.layer.cornerRadius = 8.0
        notifyLabel.layer.borderColor = UIColor.black.cgColor
        notifyLabel.layer.borderWidth = 1.0
        notifyButton.layer.cornerRadius = 8.0
        notifyButton.layer.borderColor = UIColor.black.cgColor
        notifyButton.layer.borderWidth = 1.0
        readButton.titleLabel?.adjustsFontSizeToFitWidth = true
        writeButton.titleLabel?.adjustsFontSizeToFitWidth = true
        notifyButton.titleLabel?.adjustsFontSizeToFitWidth = true
        ledButton.titleLabel?.adjustsFontSizeToFitWidth = true
        
        /* Adding observer to shift view for keyboard */
        NotificationCenter.default.addObserver(self,
                                               selector: #selector(self.keyboardNotification),
                                               name: NSNotification.Name.UIKeyboardWillChangeFrame,
                                               object: nil)
        writeField.delegate = self
        
        /* Adding observer to rotate view */
        NotificationCenter.default.addObserver(self,
                                               selector: #selector(SimpleAPViewController.viewRotated),
                                               name: NSNotification.Name.UIDeviceOrientationDidChange, object: nil)
    }

    override func didReceiveMemoryWarning()
    {
        super.didReceiveMemoryWarning()
    }
    
    // MARK: CBPeripheralDelegate
    func peripheral(_ peripheral: CBPeripheral,
                    didDiscoverServices error: Error?)
    {
        for service in peripheral.services!
        {
            if service.uuid == SAP_SERVICE_UUID
            {
                let thisService = service as CBService
                peripheral.discoverCharacteristics(nil, for: thisService)
                return
            }
        }
    }
    
    func peripheral(_ peripheral: CBPeripheral,
                    didDiscoverCharacteristicsFor service: CBService,
                    error: Error?)
    {
        for characteristic in service.characteristics!
        {
            if characteristic.uuid == SAP_LED_UUID
            {
                ledChar = characteristic
            }
            else if characteristic.uuid == SAP_WRITE_UUID
            {
                writeChar = characteristic
            }
            else if characteristic.uuid == SAP_READ_UUID
            {
                readChar = characteristic
            }
            else if characteristic.uuid == SAP_NOTIFY_UUID
            {
                notifyChar = characteristic
            }
        }
        
        if(ledChar != nil && writeChar != nil && readChar != nil && notifyChar != nil)
        {
            referencedController.timeoutTimer.invalidate()
            referencedController.connectionCon.dismiss(animated: true, completion: nil)
        }
    }
    
    /* Callback for a characteristic read/notification */
    func peripheral(_ peripheral: CBPeripheral,
                    didUpdateValueFor characteristic: CBCharacteristic,
                    error: Error?)
    {
        var values = [UInt8](repeating:0, count:characteristic.value!.count)
        var intValue: UInt32 = 0
        characteristic.value?.copyBytes(to: &values,
                                        count: characteristic.value!.count)
        memcpy(&intValue, values, 4)
        let formattedInt = String(format: "%d", intValue)
        
        if characteristic.uuid == SAP_READ_UUID
        {
            readValue.text =  "\(formattedInt)"
            readStatus.text = "Value Read"
        }
        else if characteristic.uuid == SAP_NOTIFY_UUID
        {
            notifyValue.text = "\(formattedInt)"
        }
        else
        {
            print("SAP: Invalid characteristic read (\(characteristic.uuid.uuidString))")
            return
        }
        
    }
    
    // Mark: Navigation
    @objc func keyboardNotification(notification: NSNotification)
    {
        
        if let userInfo = notification.userInfo
        {
            let endFrame = (userInfo[UIKeyboardFrameEndUserInfoKey] as? NSValue)?.cgRectValue
            let duration:TimeInterval = (userInfo[UIKeyboardAnimationDurationUserInfoKey] as? NSNumber)?.doubleValue ?? 0
            let animationCurveRawNSN = userInfo[UIKeyboardAnimationCurveUserInfoKey] as? NSNumber
            let animationCurveRaw = animationCurveRawNSN?.uintValue ?? UIViewAnimationOptions.curveEaseInOut.rawValue
            let animationCurve:UIViewAnimationOptions = UIViewAnimationOptions(rawValue: animationCurveRaw)
            if (endFrame?.origin.y)! >= UIScreen.main.bounds.size.height
            {
                bottomMargin?.constant = 10.0
            }
            else
            {
                bottomMargin?.constant = endFrame?.size.height ?? 10.0
            }
            UIView.animate(withDuration: duration,
                           delay: TimeInterval(0),
                           options: animationCurve,
                           animations: { self.view.layoutIfNeeded() },
                           completion: nil)
        }
    }
    
    @objc func viewRotated()
    {
        DispatchQueue.main.async
        {
            if UIDeviceOrientationIsPortrait(UIDevice.current.orientation)
            {
                self.readLabel.isHidden = false
                self.writeLabel.isHidden = false
                self.ledLabel.isHidden = false
                self.notifyLabel.isHidden = false
                self.writeStatus.isHidden = false
                self.notifyStatus.isHidden = false
                self.readStatus.isHidden = true
            }
            
            if UIDeviceOrientationIsLandscape(UIDevice.current.orientation)
            {
                self.readLabel.isHidden = true
                self.writeLabel.isHidden = true
                self.ledLabel.isHidden = true
                self.notifyLabel.isHidden = true
                self.writeStatus.isHidden = true
                self.notifyStatus.isHidden = true
                self.readStatus.isHidden = true
            }
        }
    }
    
    override func viewWillDisappear(_ animated: Bool)
    {
        referencedController.btManager.scanForPeripherals(withServices: nil, options: nil)
        referencedController.btManager.cancelPeripheralConnection(self.btDevice)
    }
    
    // MARK: UITextFieldDelegate
    func textFieldShouldReturn(_ textField: UITextField) -> Bool
    {
        writeField.resignFirstResponder()
        return true
    }
    
    func textFieldDidEndEditing(_ textField: UITextField)
    {
        
    }
    
    func textField(_ textField: UITextField, shouldChangeCharactersIn range: NSRange,
                   replacementString string: String) -> Bool
    {
        if((string.characters.count + (writeField.text?.characters.count)!) > 2)
        {
            return false
        }
        
        let separatedChar = string.components(separatedBy: hexSet)
        let filteredNums = separatedChar.joined(separator: "")
        
        if (filteredNums != string)
        {
            return false
        }
        
        return true
    }
}
