/*
 * Copyright (c) 2017, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

import UIKit
import CoreBluetooth
import Charts

class TemperatureSensorViewController: UIViewController, CBPeripheralDelegate
{
    // MARK: Properties
    @IBOutlet weak var tempChart: LineChartView!
    @IBOutlet weak var statisticsView: UIView!
    @IBOutlet weak var enableSwitch: UISwitch!
    @IBOutlet weak var unitSwitch: UISwitch!
    @IBOutlet weak var tempLabel: UILabel!
    @IBOutlet weak var cfLabel: UILabel!
    @IBOutlet weak var offOnLabel: UILabel!
    @IBOutlet weak var averageNameLabel: UILabel!
    @IBOutlet weak var maxNameLabel: UILabel!
    @IBOutlet weak var minNameLabel: UILabel!
    @IBOutlet weak var maxLabel: UILabel!
    @IBOutlet weak var minLabel: UILabel!
    @IBOutlet weak var averageLabel: UILabel!
    @IBOutlet weak var optionsStack: UIStackView!
    
    // MARK: Variables
    let TEMP_SERVICE_UUID = CBUUID(string: "FFE0")
    let TEMP_CHARF_UUID = CBUUID(string: "FFE2")
    let TEMP_CHARC_UUID = CBUUID(string: "FFE1")
    var tempFChar: CBCharacteristic!
    var tempCChar: CBCharacteristic!
    var btDevice: CBPeripheral!
    let maxNumOfReadings = 15
    var curIndex = 0
    var lineChartDataSet: LineChartDataSet!
    var lineChartData: LineChartData!
    var tempDataEntries: [ChartDataEntry] = []
    var simulationTimer: Timer!
    var connectingCon: DiscoveryViewController!
    var referencedController: BLEDeviceTableViewController!

    // MARK: Actions

    @IBAction func onOffChanged(_ sender: UISwitch)
    {
        if(btDevice != nil)
        {
            if sender.isOn
            {
                unitSwitch.isEnabled = true
                
                if(unitSwitch.isOn)
                {
                    btDevice.setNotifyValue(true, for: tempFChar)
                }
                else
                {
                    btDevice.setNotifyValue(true, for: tempCChar)
                }
            }
            else
            {
                unitSwitch.isEnabled = false
                btDevice.setNotifyValue(false, for: tempFChar)
                btDevice.setNotifyValue(false, for: tempCChar)
            }
        }
    }
    
    @IBAction func unitChanged(_ sender: UISwitch)
    {
        if sender.isOn
        {
            btDevice.setNotifyValue(false, for: tempCChar)
            lineChartDataSet.values.removeAll()
            curIndex = 0
            btDevice.setNotifyValue(true, for: tempFChar)
            
        }
        else
        {
            btDevice.setNotifyValue(false, for: tempFChar)
            lineChartDataSet.values.removeAll()
            curIndex = 0
            btDevice.setNotifyValue(true, for: tempCChar)
        }
    }

    
    // MARK: Functions
    override func viewDidLoad()
    {
        super.viewDidLoad()

        tempChart.rightAxis.enabled = false
        tempChart.noDataText = "Reading Data..."
        tempChart.legend.enabled = false
        tempLabel.adjustsFontSizeToFitWidth = true
        
        averageNameLabel.adjustsFontSizeToFitWidth = true
        maxNameLabel.adjustsFontSizeToFitWidth = true
        minNameLabel.adjustsFontSizeToFitWidth = true
        offOnLabel.adjustsFontSizeToFitWidth = true
        cfLabel.adjustsFontSizeToFitWidth = true
        statisticsView.layer.borderColor = UIColor.black.cgColor
        statisticsView.layer.borderWidth = 3.0
        statisticsView.layer.cornerRadius = 8.0
        statisticsView.clipsToBounds = false
        
        /* Adding observer to rotate view */
        NotificationCenter.default.addObserver(self,
                                               selector: #selector(TemperatureSensorViewController.viewRotated),
                                               name: NSNotification.Name.UIDeviceOrientationDidChange, object: nil)
        
    }

    override func didReceiveMemoryWarning()
    {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }

    // MARK: CBPeripheralDelegate
    func peripheral(_ peripheral: CBPeripheral,
                    didDiscoverServices error: Error?)
    {
        for service in peripheral.services!
        {
            if service.uuid == TEMP_SERVICE_UUID
            {
                let thisService = service as CBService
                peripheral.discoverCharacteristics(nil, for: thisService)
                return
            }
        }
    }
    
    func peripheral(_ peripheral: CBPeripheral,
                    didDiscoverCharacteristicsFor service: CBService,
                    error: Error?)
    {
        
        for characteristic in service.characteristics!
        {
            if characteristic.uuid == TEMP_CHARF_UUID
            {
                tempFChar = characteristic
            }
            else if characteristic.uuid == TEMP_CHARC_UUID
            {
                tempCChar = characteristic
            }
            
            if tempFChar != nil && tempCChar != nil
            {
                btDevice.setNotifyValue(true, for: tempFChar)
                referencedController.timeoutTimer.invalidate()
                referencedController.connectionCon.dismiss(animated: true, completion: nil)
            }
            
        }
    }
    
    /* Callback for a characteristic read/notification */
    func peripheral(_ peripheral: CBPeripheral,
                    didUpdateValueFor characteristic: CBCharacteristic,
                    error: Error?)
    {
        var values = [UInt8](repeating:0, count:characteristic.value!.count)
        var floatValue: Float = 0.0
        characteristic.value?.copyBytes(to: &values,
                                        count: characteristic.value!.count)
        memcpy(&floatValue, values, 4)
        let formattedFloat = String(format: "%.1f", floatValue)
        
        if characteristic.uuid == TEMP_CHARF_UUID
        {
            print("Temperature notification read: \(formattedFloat)°F")
            tempLabel.text =  "\(formattedFloat)°F"
            
        }
        else if characteristic.uuid == TEMP_CHARC_UUID
        {
            print("Temperature notification read: \(formattedFloat)°C")
            tempLabel.text =  "\(formattedFloat)°C"
        }
        else
        {
            print("Themometer: Invalid characteristic read (\(characteristic.uuid.uuidString))")
            return
        }
        
        if (curIndex == 0)
        {
            lineChartDataSet = LineChartDataSet(values: tempDataEntries,
                                                label: "Temperature")
            adjustChartFormat()
        }
        
        let dataEntry = ChartDataEntry(x: Double(curIndex),
                                       y: Double(floatValue))
        
        if(curIndex == (maxNumOfReadings-1))
        {
            _ = lineChartDataSet.removeFirst()
            
            for dataValue in lineChartDataSet.values
            {
                dataValue.x = dataValue.x - 1
            }
            
        }
        else
        {
            curIndex = curIndex + 1
        }
        
        _ = lineChartDataSet.addEntry(dataEntry)
        lineChartData.notifyDataChanged()
        tempChart.notifyDataSetChanged()
        
        calculateChartStatistics()
        
    }
    
    func adjustChartFormat()
    {
        let gradientColors = [UIColor(red: 0.5, green: 0.0, blue: 0.0, alpha: 0.7).cgColor,
                              UIColor(red: 1.0, green: 0.0, blue: 0.0, alpha: 0.7).cgColor]
        let colorLocations: [CGFloat] = [1.0, 0.0]
        let gradient = CGGradient(colorsSpace: CGColorSpaceCreateDeviceRGB(),
                                  colors: gradientColors as CFArray,
                                  locations: colorLocations)
        lineChartDataSet.fill = Fill.fillWithLinearGradient(gradient!,
                                                            angle: 90.0)
        lineChartDataSet.setCircleColor(UIColor.black)
        lineChartDataSet.circleRadius = 4.0
        lineChartDataSet.drawCircleHoleEnabled = false
        lineChartDataSet.fillAlpha = 1.0
        lineChartDataSet.drawFilledEnabled = true
        lineChartData = LineChartData()
        lineChartData.addDataSet(lineChartDataSet)
        tempChart.data = lineChartData
        
        /* Adjusting the Margins */
        let yAxis = tempChart.leftAxis
        
        if(unitSwitch.isOn)
        {
            yAxis.axisMaximum = 100
            yAxis.axisMinimum = 30
        }
        else
        {
            yAxis.axisMaximum = 42
            yAxis.axisMinimum = -5
        }
        
        let xAxis = tempChart.xAxis
        xAxis.enabled = false
        
        if btDevice != nil
        {
            tempChart.chartDescription?.text = "Temperature Data from MSP432"
        }
        else
        {
            tempChart.chartDescription?.text = "SIMULATION MODE ONLY!"
        }
        
    }
    
    func calculateChartStatistics()
    {
        var curMin = lineChartDataSet.values[0].y
        var curMax = curMin
        var runningSum = 0.0
        
        for dataValue in lineChartDataSet.values
        {
            if(dataValue.y > curMax)
            {
                curMax = dataValue.y
            }
            
            if(dataValue.y < curMin)
            {
                curMin = dataValue.y
            }
            
            runningSum = runningSum + dataValue.y
        }
        
        let formattedMin = String(format: "%.1f", curMin)
        let formattedMax = String(format: "%.1f", curMax)
        let formattedAverage = String(format: "%.1f",
                                      (runningSum/(Double(curIndex))))
        
        maxLabel.text = formattedMax
        minLabel.text = formattedMin
        averageLabel.text = formattedAverage
    }

    override func viewWillDisappear(_ animated: Bool)
    {
        referencedController.btManager.scanForPeripherals(withServices: nil, options: nil)
        referencedController.btManager.cancelPeripheralConnection(self.btDevice)
    }
    
    @objc func viewRotated()
    {
        DispatchQueue.main.async
            {
                if UIDeviceOrientationIsPortrait(UIDevice.current.orientation)
                {
                    self.optionsStack.isHidden = false
                    self.statisticsView.isHidden = false
                }
                
                if UIDeviceOrientationIsLandscape(UIDevice.current.orientation)
                {
                    self.statisticsView.isHidden = true
                    self.optionsStack.isHidden = true
                }
        }
    }
    
}
